from __future__ import annotations

from functools import partial
from itertools import chain, combinations, repeat
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import seaborn as sns
from anndata import AnnData
from matplotlib.testing.compare import compare_images
from packaging.version import Version

import scanpy as sc
from scanpy._compat import pkg_version
from testing.scanpy._helpers.data import (
    krumsiek11,
    pbmc3k,
    pbmc3k_processed,
    pbmc68k_reduced,
)
from testing.scanpy._pytest.marks import needs

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

    from matplotlib.axes import Axes


HERE: Path = Path(__file__).parent
ROOT = HERE / "_images"


# Test images are saved in the directory ./_images/<test-name>/
# If test images need to be updated, simply copy actual.png to expected.png.


@pytest.mark.parametrize("col", [None, "symb"])
@pytest.mark.parametrize("layer", [None, "layer_name"])
def test_highest_expr_genes(image_comparer, col, layer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=5)

    adata = pbmc3k()
    if layer is not None:
        adata.layers[layer] = adata.X
        del adata.X
    # check that only existing categories are shown
    adata.var["symb"] = adata.var_names.astype("category")

    sc.pl.highest_expr_genes(adata, 20, gene_symbols=col, layer=layer, show=False)

    save_and_compare_images("highest_expr_genes")


@needs.leidenalg
@pytest.mark.parametrize(
    ("params", "key"),
    [
        pytest.param({}, "heatmap", id="default"),
        pytest.param(
            dict(swap_axes=True, figsize=(10, 3), cmap="YlGnBu"),
            "heatmap_swap_axes",
            id="swap",
        ),
        pytest.param(
            dict(
                groupby="numeric_value",
                num_categories=4,
                figsize=(4.5, 5),
                dendrogram=False,
            ),
            "heatmap2",
            id="numeric",
        ),
        pytest.param(
            dict(standard_scale="var", layer="test"),
            "heatmap_std_scale_var",
            id="std_scale=var",
        ),
        pytest.param(
            dict(standard_scale="obs"),
            "heatmap_std_scale_obs",
            id="std_scale=obs",
        ),
    ],
)
def test_heatmap(image_comparer, params: dict[str, Any], key: str) -> None:
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = krumsiek11()
    adata.obs["numeric_value"] = adata.X[:, 0]
    adata.layers["test"] = -1 * adata.X.copy()

    params = dict(groupby="cell_type", dendrogram=True) | params
    sc.pl.heatmap(adata, adata.var_names, **params, use_raw=False, show=False)
    save_and_compare_images(key)


@needs.leidenalg
def test_heatmap_var_as_dict(image_comparer) -> None:
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.tl.leiden(
        pbmc,
        key_added="clusters",
        resolution=0.5,
        flavor="igraph",
        n_iterations=2,
        directed=False,
    )
    # call umap to trigger colors for the clusters
    sc.pl.umap(pbmc, color="clusters")
    marker_genes_dict = {
        "3": ["GNLY", "NKG7"],
        "1": ["FCER1A"],
        "2": ["CD3D"],
        "0": ["FCGR3A"],
        "4": ["CD79A", "MS4A1"],
    }
    sc.pl.heatmap(
        adata=pbmc,
        var_names=marker_genes_dict,
        groupby="clusters",
        vmin=-2,
        vmax=2,
        cmap="RdBu_r",
        dendrogram=True,
        swap_axes=True,
    )
    save_and_compare_images("heatmap_var_as_dict")


@needs.leidenalg
@pytest.mark.parametrize("swap_axes", [True, False])
def test_heatmap_alignment(*, image_comparer, swap_axes: bool) -> None:
    """Test that plot elements are well aligned."""
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    a = AnnData(
        np.array([[0, 0.3, 0.5], [1, 1.3, 1.5], [2, 2.3, 2.5]]),
        obs={"foo": ["a", "b", "c"]},
        var=pd.DataFrame({"genes": ["g1", "g2", "g3"]}).set_index("genes"),
    )
    a.obs["foo"] = a.obs["foo"].astype("category")
    sc.pl.heatmap(
        a, var_names=a.var_names, groupby="foo", swap_axes=swap_axes, figsize=(4, 4)
    )
    save_and_compare_images(f"heatmap_small{'_swap' if swap_axes else ''}_alignment")


@pytest.mark.parametrize(
    ("obs_keys", "name"),
    [(None, "clustermap"), ("cell_type", "clustermap_withcolor")],
)
def test_clustermap(image_comparer, obs_keys, name):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = krumsiek11()
    sc.pl.clustermap(adata, obs_keys)
    save_and_compare_images(name)


params_dotplot_matrixplot_stacked_violin = [
    pytest.param(id, fn, id=id)
    for id, fn in [
        (
            "dotplot",
            partial(
                sc.pl.dotplot, groupby="cell_type", title="dotplot", dendrogram=True
            ),
        ),
        (
            "dotplot2",
            partial(
                sc.pl.dotplot,
                groupby="numeric_column",
                use_raw=False,
                num_categories=7,
                title="non categorical obs",
                figsize=(7, 2.5),
            ),
        ),
        (
            "dotplot3",
            partial(
                sc.pl.dotplot,
                groupby="cell_type",
                dot_max=0.7,
                dot_min=0.1,
                cmap="hot_r",
                title="dot_max=0.7 dot_min=0.1, var_groups",
                var_group_positions=[(0, 1), (9, 10)],
                var_group_labels=["A", "B"],
                dendrogram=True,
            ),
        ),
        (
            "dotplot_std_scale_group",
            partial(
                sc.pl.dotplot,
                groupby="cell_type",
                use_raw=False,
                dendrogram=True,
                layer="test",
                swap_axes=True,
                title="swap_axes, layer=-1*X, scale=group\nsmallest_dot=10",
                standard_scale="group",
                smallest_dot=10,
            ),
        ),
        (
            "dotplot_dict",
            partial(
                sc.pl.dotplot,
                groupby="cell_type",
                dot_max=0.7,
                dot_min=0.1,
                color_map="winter",
                title="var as dict",
                dendrogram=True,
            ),
        ),
        (
            "matrixplot",
            partial(
                sc.pl.matrixplot,
                groupby="cell_type",
                use_raw=False,
                title="matrixplot",
                dendrogram=True,
            ),
        ),
        (
            "matrixplot_std_scale_var_dict",
            partial(
                sc.pl.matrixplot,
                groupby="cell_type",
                dendrogram=True,
                standard_scale="var",
                layer="test",
                cmap="Blues_r",
                title='scale var, custom colorbar_title, layer="test"',
                colorbar_title="Scaled expression",
            ),
        ),
        (
            "matrixplot_std_scale_group",
            partial(
                sc.pl.matrixplot,
                groupby="cell_type",
                use_raw=False,
                standard_scale="group",
                title="scale_group, swap_axes",
                swap_axes=True,
            ),
        ),
        (
            "matrixplot2",
            partial(
                sc.pl.matrixplot,
                groupby="numeric_column",
                use_raw=False,
                num_categories=4,
                title="non-categorical obs, custom figsize",
                figsize=(8, 2.5),
                cmap="RdBu_r",
            ),
        ),
        (
            "stacked_violin",
            partial(
                sc.pl.stacked_violin,
                groupby="cell_type",
                use_raw=False,
                title="stacked_violin",
                dendrogram=True,
            ),
        ),
        (
            "stacked_violin_std_scale_var_dict",
            partial(
                sc.pl.stacked_violin,
                groupby="cell_type",
                dendrogram=True,
                standard_scale="var",
                layer="test",
                title='scale var, layer="test"',
            ),
        ),
        (
            "stacked_violin_std_scale_group",
            partial(
                sc.pl.stacked_violin,
                groupby="cell_type",
                use_raw=False,
                standard_scale="group",
                title="scale_group\nswap_axes",
                swap_axes=True,
                cmap="Blues",
            ),
        ),
        (
            "stacked_violin_no_cat_obs",
            partial(
                sc.pl.stacked_violin,
                groupby="numeric_column",
                use_raw=False,
                num_categories=4,
                title="non-categorical obs, custom figsize",
                figsize=(8, 2.5),
            ),
        ),
    ]
]


@pytest.mark.parametrize(("id", "fn"), params_dotplot_matrixplot_stacked_violin)
def test_dotplot_matrixplot_stacked_violin(image_comparer, id, fn):
    save_and_compare_images = partial(image_comparer, ROOT, tol=5)

    adata = krumsiek11()
    adata.obs["numeric_column"] = adata.X[:, 0]
    adata.layers["test"] = -1 * adata.X.copy()
    genes_dict = {
        "group a": ["Gata2", "Gata1"],
        "group b": ["Fog1", "EKLF", "Fli1", "SCL"],
        "group c": ["Cebpa", "Pu.1", "cJun", "EgrNab", "Gfi1"],
    }

    if id.endswith("dict"):
        fn(adata, genes_dict, show=False)
    else:
        fn(adata, adata.var_names, show=False)
    save_and_compare_images(id)


def test_dotplot_obj(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    # test dotplot dot_min, dot_max, color_map, and var_groups
    pbmc = pbmc68k_reduced()
    genes = [
        *["CD79A", "MS4A1", "CD8A", "CD8B", "LYZ", "LGALS3"],
        *["S100A8", "GNLY", "NKG7", "KLRB1", "FCGR3A", "FCER1A", "CST3"],
    ]
    # test layer, var standardization, smallest_dot,
    # color title, size_title return_fig and dot_edge
    pbmc.layers["test"] = pbmc.X * -1
    plot = sc.pl.dotplot(
        pbmc,
        genes,
        "bulk_labels",
        layer="test",
        dendrogram=True,
        return_fig=True,
        standard_scale="var",
        smallest_dot=40,
        colorbar_title="scaled column max",
        size_title="Fraction of cells",
    )
    plot.style(dot_edge_color="black", dot_edge_lw=0.1, cmap="Reds").show()

    save_and_compare_images("dotplot_std_scale_var")


def test_dotplot_style_no_reset():
    pbmc = pbmc68k_reduced()
    plot = sc.pl.dotplot(pbmc, "CD79A", "bulk_labels", return_fig=True)
    assert isinstance(plot, sc.pl.DotPlot)
    assert plot.cmap == sc.pl.DotPlot.DEFAULT_COLORMAP
    plot.style(cmap="winter")
    assert plot.cmap == "winter"
    plot.style(color_on="square")
    assert plot.cmap == "winter", "style() should not reset unspecified parameters"


def test_dotplot_add_totals(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=5)

    pbmc = pbmc68k_reduced()
    markers = {"T-cell": "CD3D", "B-cell": "CD79A", "myeloid": "CST3"}
    sc.pl.dotplot(pbmc, markers, "bulk_labels", return_fig=True).add_totals().show()
    save_and_compare_images("dotplot_totals")


def test_matrixplot_obj(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = pbmc68k_reduced()
    marker_genes_dict = {
        "3": ["GNLY", "NKG7"],
        "1": ["FCER1A"],
        "2": ["CD3D"],
        "0": ["FCGR3A"],
        "4": ["CD79A", "MS4A1"],
    }

    plot = sc.pl.matrixplot(
        adata,
        marker_genes_dict,
        "bulk_labels",
        use_raw=False,
        title="added totals",
        return_fig=True,
    )
    plot.add_totals(sort="descending").style(edge_color="white", edge_lw=0.5).show()
    save_and_compare_images("matrixplot_with_totals")

    axes = plot.get_axes()
    assert "mainplot_ax" in axes, "mainplot_ax not found in returned axes dict"


def test_stacked_violin_obj(image_comparer, plt):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    markers = {
        "T-cell": ["CD3D", "CD3E", "IL32"],
        "B-cell": ["CD79A", "CD79B", "MS4A1"],
        "myeloid": ["CST3", "LYZ"],
    }
    plot = sc.pl.stacked_violin(
        pbmc,
        markers,
        "bulk_labels",
        use_raw=False,
        title="return_fig. add_totals",
        return_fig=True,
    )
    plot.add_totals().style(row_palette="tab20").show()
    save_and_compare_images("stacked_violin_return_fig")


# checking for https://github.com/scverse/scanpy/issues/3152
def test_stacked_violin_swap_axes_match(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=10)
    pbmc = pbmc68k_reduced()
    sc.tl.rank_genes_groups(
        pbmc,
        "bulk_labels",
        method="wilcoxon",
        tie_correct=True,
        pts=True,
        key_added="wilcoxon",
    )
    swapped_ax = sc.pl.rank_genes_groups_stacked_violin(
        pbmc,
        n_genes=2,
        key="wilcoxon",
        groupby="bulk_labels",
        swap_axes=True,
        return_fig=True,
    )
    swapped_ax.show()
    save_and_compare_images("stacked_violin_swap_axes_pbmc68k_reduced")


def test_tracksplot(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = krumsiek11()
    sc.pl.tracksplot(
        adata, adata.var_names, "cell_type", dendrogram=True, use_raw=False
    )
    save_and_compare_images("tracksplot")


def test_multiple_plots(image_comparer):
    # only testing stacked_violin, matrixplot and dotplot
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = pbmc68k_reduced()
    markers = {
        "T-cell": ["CD3D", "CD3E", "IL32"],
        "B-cell": ["CD79A", "CD79B", "MS4A1"],
        "myeloid": ["CST3", "LYZ"],
    }
    _fig, (ax1, ax2, ax3) = plt.subplots(
        1, 3, figsize=(20, 5), gridspec_kw={"wspace": 0.7}
    )
    _ = sc.pl.stacked_violin(
        adata,
        markers,
        groupby="bulk_labels",
        ax=ax1,
        title="stacked_violin",
        dendrogram=True,
        show=False,
    )
    _ = sc.pl.dotplot(
        adata,
        markers,
        groupby="bulk_labels",
        ax=ax2,
        title="dotplot",
        dendrogram=True,
        show=False,
    )
    _ = sc.pl.matrixplot(
        adata,
        markers,
        groupby="bulk_labels",
        ax=ax3,
        title="matrixplot",
        dendrogram=True,
        show=False,
    )
    save_and_compare_images("multiple_plots")


def test_violin(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=40)

    with plt.rc_context():
        sc.pl.set_rcParams_defaults()
        sc.set_figure_params(dpi=50, color_map="viridis")

        pbmc = pbmc68k_reduced()
        sc.pl.violin(
            pbmc,
            ["n_genes", "percent_mito", "n_counts"],
            stripplot=True,
            multi_panel=True,
            jitter=True,
            show=False,
        )
        save_and_compare_images("violin_multi_panel")

        sc.pl.violin(
            pbmc,
            ["n_genes", "percent_mito", "n_counts"],
            ylabel=["foo", "bar", "baz"],
            groupby="bulk_labels",
            stripplot=True,
            multi_panel=True,
            jitter=True,
            show=False,
            rotation=90,
        )
        save_and_compare_images("violin_multi_panel_with_groupby")

        # test use of layer
        pbmc.layers["negative"] = pbmc.X * -1
        sc.pl.violin(
            pbmc,
            "CST3",
            groupby="bulk_labels",
            stripplot=True,
            multi_panel=True,
            jitter=True,
            show=False,
            layer="negative",
            use_raw=False,
            rotation=90,
        )
        save_and_compare_images("violin_multi_panel_with_layer")


# TODO: Generalize test to more plotting types
def test_violin_without_raw(tmp_path):
    # https://github.com/scverse/scanpy/issues/1546
    has_raw_pth = tmp_path / "has_raw.png"
    no_raw_pth = tmp_path / "no_raw.png"

    pbmc = pbmc68k_reduced()
    pbmc_no_raw = pbmc.raw.to_adata().copy()

    sc.pl.violin(pbmc, "CST3", groupby="bulk_labels", show=False, jitter=False)
    plt.savefig(has_raw_pth)
    plt.close()

    sc.pl.violin(pbmc_no_raw, "CST3", groupby="bulk_labels", show=False, jitter=False)
    plt.savefig(no_raw_pth)
    plt.close()

    assert compare_images(has_raw_pth, no_raw_pth, tol=5) is None


def test_dendrogram(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=10)

    pbmc = pbmc68k_reduced()
    sc.pl.dendrogram(pbmc, "bulk_labels")
    save_and_compare_images("dendrogram")


def test_correlation(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.pl.correlation_matrix(pbmc, "bulk_labels")
    save_and_compare_images("correlation")


_RANK_GENES_GROUPS_PARAMS = [
    (
        "sharey",
        partial(sc.pl.rank_genes_groups, n_genes=12, n_panels_per_row=3, show=False),
    ),
    (
        "basic",
        partial(
            sc.pl.rank_genes_groups,
            n_genes=12,
            n_panels_per_row=3,
            sharey=False,
            show=False,
        ),
    ),
    (
        "heatmap",
        partial(sc.pl.rank_genes_groups_heatmap, n_genes=4, cmap="YlGnBu", show=False),
    ),
    (
        "heatmap_swap_axes",
        partial(
            sc.pl.rank_genes_groups_heatmap,
            n_genes=20,
            swap_axes=True,
            use_raw=False,
            show_gene_labels=False,
            show=False,
            vmin=-3,
            vmax=3,
            cmap="bwr",
        ),
    ),
    (
        "heatmap_swap_axes_vcenter",
        partial(
            sc.pl.rank_genes_groups_heatmap,
            n_genes=20,
            swap_axes=True,
            use_raw=False,
            show_gene_labels=False,
            show=False,
            vmin=-3,
            vcenter=1,
            vmax=3,
            cmap="RdBu_r",
        ),
    ),
    (
        "stacked_violin",
        partial(
            sc.pl.rank_genes_groups_stacked_violin,
            n_genes=3,
            show=False,
            groups=["3", "0", "5"],
        ),
    ),
    (
        "dotplot",
        partial(sc.pl.rank_genes_groups_dotplot, n_genes=4, show=False),
    ),
    (
        "dotplot_gene_names",
        partial(
            sc.pl.rank_genes_groups_dotplot,
            var_names={
                "T-cell": ["CD3D", "CD3E", "IL32"],
                "B-cell": ["CD79A", "CD79B", "MS4A1"],
                "myeloid": ["CST3", "LYZ"],
            },
            values_to_plot="logfoldchanges",
            cmap="bwr",
            vmin=-3,
            vmax=3,
            show=False,
        ),
    ),
    (
        "dotplot_logfoldchange",
        partial(
            sc.pl.rank_genes_groups_dotplot,
            n_genes=4,
            values_to_plot="logfoldchanges",
            vmin=-5,
            vmax=5,
            min_logfoldchange=3,
            cmap="RdBu_r",
            swap_axes=True,
            title="log fold changes swap_axes",
            show=False,
        ),
    ),
    (
        "dotplot_logfoldchange_vcenter",
        partial(
            sc.pl.rank_genes_groups_dotplot,
            n_genes=4,
            values_to_plot="logfoldchanges",
            vmin=-5,
            vcenter=1,
            vmax=5,
            min_logfoldchange=3,
            cmap="RdBu_r",
            swap_axes=True,
            title="log fold changes swap_axes",
            show=False,
        ),
    ),
    (
        "matrixplot",
        partial(
            sc.pl.rank_genes_groups_matrixplot,
            n_genes=5,
            show=False,
            title="matrixplot",
            gene_symbols="symbol",
            use_raw=False,
        ),
    ),
    (
        "matrixplot_gene_names_symbol",
        partial(
            sc.pl.rank_genes_groups_matrixplot,
            var_names={
                "T-cell": ["CD3D__", "CD3E__", "IL32__"],
                "B-cell": ["CD79A__", "CD79B__", "MS4A1__"],
                "myeloid": ["CST3__", "LYZ__"],
            },
            values_to_plot="logfoldchanges",
            cmap="bwr",
            vmin=-3,
            vmax=3,
            gene_symbols="symbol",
            use_raw=False,
            show=False,
        ),
    ),
    (
        "matrixplot_n_genes_negative",
        partial(
            sc.pl.rank_genes_groups_matrixplot,
            n_genes=-5,
            show=False,
            title="matrixplot n_genes=-5",
        ),
    ),
    (
        "matrixplot_swap_axes",
        partial(
            sc.pl.rank_genes_groups_matrixplot,
            n_genes=5,
            show=False,
            swap_axes=True,
            values_to_plot="logfoldchanges",
            vmin=-6,
            vmax=6,
            cmap="bwr",
            title="log fold changes swap_axes",
        ),
    ),
    (
        "matrixplot_swap_axes_vcenter",
        partial(
            sc.pl.rank_genes_groups_matrixplot,
            n_genes=5,
            show=False,
            swap_axes=True,
            values_to_plot="logfoldchanges",
            vmin=-6,
            vcenter=1,
            vmax=6,
            cmap="bwr",
            title="log fold changes swap_axes",
        ),
    ),
    (
        "tracksplot",
        partial(
            sc.pl.rank_genes_groups_tracksplot,
            n_genes=3,
            show=False,
            groups=["3", "2", "1"],
        ),
    ),
    (
        "violin",
        partial(
            sc.pl.rank_genes_groups_violin,
            groups="0",
            n_genes=5,
            use_raw=True,
            jitter=False,
            strip=False,
            show=False,
        ),
    ),
    (
        "violin_not_raw",
        partial(
            sc.pl.rank_genes_groups_violin,
            groups="0",
            n_genes=5,
            use_raw=False,
            jitter=False,
            strip=False,
            show=False,
        ),
    ),
]


@pytest.mark.parametrize(
    ("name", "fn"),
    [pytest.param(name, fn, id=name) for name, fn in _RANK_GENES_GROUPS_PARAMS],
)
def test_rank_genes_groups(image_comparer, name, fn):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.tl.rank_genes_groups(pbmc, "louvain", n_genes=pbmc.raw.shape[1])

    # add gene symbol
    pbmc.var["symbol"] = pbmc.var.index + "__"

    with plt.rc_context({"axes.grid": True, "figure.figsize": (4, 4)}):
        fn(pbmc)
    key = "ranked_genes" if name == "basic" else f"ranked_genes_{name}"
    save_and_compare_images(key)
    plt.close()


def test_rank_genes_group_axes(image_comparer):
    fn = next(fn for name, fn in _RANK_GENES_GROUPS_PARAMS if name == "basic")

    save_and_compare_images = partial(image_comparer, ROOT, tol=23)

    pbmc = pbmc68k_reduced()
    sc.tl.rank_genes_groups(pbmc, "louvain", n_genes=pbmc.raw.shape[1])

    pbmc.var["symbol"] = pbmc.var.index + "__"

    _fig, ax = plt.subplots(figsize=(12, 16))
    ax.set_axis_off()
    with plt.rc_context({"axes.grid": True}):
        axes: list[Axes] = fn(pbmc, ax=ax, show=False)

    assert len(axes) == 11
    _fig.show()
    save_and_compare_images("ranked_genes")
    plt.close()


@pytest.fixture(scope="session")
def gene_symbols_adatas_session() -> tuple[AnnData, AnnData]:
    """Create two anndata objects which are equivalent except for var_names.

    Both have ensembl ids and hgnc symbols as columns in var. The first has ensembl
    ids as var_names, the second has symbols.
    """
    pbmc = pbmc3k_processed().raw.to_adata()
    pbmc_counts = pbmc3k()

    pbmc.layers["counts"] = pbmc_counts[pbmc.obs_names, pbmc.var_names].X.copy()
    pbmc.var["gene_symbol"] = pbmc.var_names
    pbmc.var["ensembl_id"] = pbmc_counts.var["gene_ids"].loc[pbmc.var_names]

    pbmc.var = pbmc.var.set_index("ensembl_id", drop=False)

    # Cutting down on size for plotting, tracksplot and stacked_violin are slow
    pbmc = pbmc[pbmc.obs["louvain"].isin(pbmc.obs["louvain"].cat.categories[:4])]
    pbmc = pbmc[::3].copy()

    # Creating variations
    a = pbmc.copy()
    b = pbmc.copy()
    a.var = a.var.set_index("ensembl_id")
    b.var = b.var.set_index("gene_symbol")

    # Computing DE
    sc.tl.rank_genes_groups(a, groupby="louvain")
    sc.tl.rank_genes_groups(b, groupby="louvain")

    return a, b


@pytest.fixture
def gene_symbols_adatas(gene_symbols_adatas_session) -> tuple[AnnData, AnnData]:
    a, b = gene_symbols_adatas_session
    return a.copy(), b.copy()


@pytest.mark.parametrize(
    "func",
    [
        sc.pl.rank_genes_groups_dotplot,
        sc.pl.rank_genes_groups_heatmap,
        sc.pl.rank_genes_groups_matrixplot,
        sc.pl.rank_genes_groups_stacked_violin,
        sc.pl.rank_genes_groups_tracksplot,
        # TODO: add other rank_genes_groups plots here once they work
    ],
)
def test_plot_rank_genes_groups_gene_symbols(
    gene_symbols_adatas, func, tmp_path, check_same_image
):
    a, b = gene_symbols_adatas

    pth_1_a = tmp_path / f"{func.__name__}_equivalent_gene_symbols_1_a.png"
    pth_1_b = tmp_path / f"{func.__name__}_equivalent_gene_symbols_1_b.png"

    func(a, gene_symbols="gene_symbol")
    plt.savefig(pth_1_a)
    plt.close()

    func(b)
    plt.savefig(pth_1_b)

    check_same_image(pth_1_a, pth_1_b, tol=1, root=tmp_path)

    pth_2_a = tmp_path / f"{func.__name__}_equivalent_gene_symbols_2_a.png"
    pth_2_b = tmp_path / f"{func.__name__}_equivalent_gene_symbols_2_b.png"

    func(a)
    plt.savefig(pth_2_a)
    plt.close()

    func(b, gene_symbols="ensembl_id")
    plt.savefig(pth_2_b)
    plt.close()

    check_same_image(pth_2_a, pth_2_b, tol=1, root=tmp_path)


@pytest.mark.parametrize(
    "func",
    [
        sc.pl.rank_genes_groups_dotplot,
        sc.pl.rank_genes_groups_heatmap,
        sc.pl.rank_genes_groups_matrixplot,
        sc.pl.rank_genes_groups_stacked_violin,
        sc.pl.rank_genes_groups_tracksplot,
        # TODO: add other rank_genes_groups plots here once they work
    ],
)
def test_rank_genes_groups_plots_n_genes_vs_var_names(tmp_path, func, check_same_image):
    """Checks that once can pass a negative value for n_genes and var_names as a dict."""
    n = 3
    pbmc = pbmc68k_reduced().raw.to_adata()
    groups = pbmc.obs["louvain"].cat.categories[:3]
    pbmc = pbmc[pbmc.obs["louvain"].isin(groups)][::3].copy()

    sc.tl.rank_genes_groups(pbmc, groupby="louvain")

    top_genes = {}
    bottom_genes = {}
    for g, subdf in sc.get.rank_genes_groups_df(pbmc, group=groups).groupby(
        "group", observed=True
    ):
        top_genes[g] = list(subdf["names"].head(n))
        bottom_genes[g] = list(subdf["names"].tail(n))

    positive_n_pth = tmp_path / f"{func.__name__}_positive_n.png"
    top_genes_pth = tmp_path / f"{func.__name__}_top_genes.png"
    negative_n_pth = tmp_path / f"{func.__name__}_negative_n.png"
    bottom_genes_pth = tmp_path / f"{func.__name__}_bottom_genes.png"

    def wrapped(pth, **kwargs):
        func(pbmc, groupby="louvain", dendrogram=False, **kwargs)
        plt.savefig(pth)
        plt.close()

    wrapped(positive_n_pth, n_genes=n)
    wrapped(top_genes_pth, var_names=top_genes)

    check_same_image(positive_n_pth, top_genes_pth, tol=1, root=tmp_path)

    wrapped(negative_n_pth, n_genes=-n)
    wrapped(bottom_genes_pth, var_names=bottom_genes)

    check_same_image(negative_n_pth, bottom_genes_pth, tol=1, root=tmp_path)

    # Shouldn't be able to pass these together
    with pytest.raises(
        ValueError, match="n_genes and var_names are mutually exclusive"
    ):
        wrapped(tmp_path / "not_written.png", n_genes=n, var_names=top_genes)


@pytest.mark.parametrize(
    ("id", "fn"),
    [
        ("heatmap", sc.pl.heatmap),
        ("dotplot", sc.pl.dotplot),
        ("matrixplot", sc.pl.matrixplot),
        ("stacked_violin", sc.pl.stacked_violin),
        ("tracksplot", sc.pl.tracksplot),
    ],
)
def test_genes_symbols(image_comparer, id, fn):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    adata = krumsiek11()

    # add a 'symbols' column
    adata.var["symbols"] = adata.var.index.map(lambda x: f"symbol_{x}")
    symbols = [f"symbol_{x}" for x in adata.var_names]

    fn(adata, symbols, "cell_type", dendrogram=True, gene_symbols="symbols", show=False)
    save_and_compare_images(f"{id}_gene_symbols")


@pytest.fixture(scope="session")
def pbmc_scatterplots_session() -> AnnData:
    # Wrapped in another fixture to avoid mutation
    pbmc = pbmc68k_reduced()
    pbmc.obs["mask"] = pbmc.obs["louvain"].isin(["0", "1", "3"])
    pbmc.layers["sparse"] = pbmc.raw.X / 2
    pbmc.layers["test"] = pbmc.X.copy() + 100
    pbmc.var["numbers"] = [str(x) for x in range(pbmc.shape[1])]
    sc.pp.neighbors(pbmc)
    sc.tl.tsne(pbmc, random_state=0, n_pcs=30)
    sc.tl.diffmap(pbmc)
    return pbmc


@pytest.fixture
def pbmc_scatterplots(pbmc_scatterplots_session) -> AnnData:
    return pbmc_scatterplots_session.copy()


@pytest.mark.parametrize(
    ("id", "fn"),
    [
        ("pca", partial(sc.pl.pca, color="bulk_labels")),
        (
            "pca_with_fonts",
            partial(
                sc.pl.pca,
                color=["bulk_labels", "louvain"],
                legend_loc="on data",
                legend_fontoutline=2,
                legend_fontweight="normal",
                legend_fontsize=10,
            ),
        ),
        pytest.param(
            "3dprojection", partial(sc.pl.pca, color="bulk_labels", projection="3d")
        ),
        (
            "multipanel",
            partial(
                sc.pl.pca,
                color=["CD3D", "CD79A"],
                components=["1,2", "1,3"],
                vmax=5,
                use_raw=False,
                vmin=-5,
                cmap="seismic",
            ),
        ),
        (
            "multipanel_vcenter",
            partial(
                sc.pl.pca,
                color=["CD3D", "CD79A"],
                components=["1,2", "1,3"],
                vmax=5,
                use_raw=False,
                vmin=-5,
                vcenter=1,
                cmap="seismic",
            ),
        ),
        (
            "pca_one_marker",
            partial(sc.pl.pca, color="louvain", marker="^"),
        ),
        (
            "pca_one_marker_multiple_colors",
            partial(sc.pl.pca, color=["louvain", "bulk_labels"], marker="^"),
        ),
        (
            "pca_multiple_markers_multiple_colors",
            partial(sc.pl.pca, color=["louvain", "bulk_labels"], marker=["^", "x"]),
        ),
        (
            "pca_marker_with_dimensions",
            partial(
                sc.pl.pca, color="louvain", marker="^", dimensions=[(0, 1), (1, 2)]
            ),
        ),
        (
            "pca_markers_with_dimensions",
            partial(
                sc.pl.pca,
                color="louvain",
                marker=["^", "x"],
                dimensions=[(0, 1), (1, 2)],
            ),
        ),
        (
            "pca_markers_colors_with_dimensions",
            partial(
                sc.pl.pca,
                color=["louvain", "bulk_labels"],
                marker=["^", "x"],
                dimensions=[(0, 1), (1, 2)],
            ),
        ),
        (
            "pca_sparse_layer",
            partial(sc.pl.pca, color=["CD3D", "CD79A"], layer="sparse", cmap="viridis"),
        ),
        # pytest.param(
        #     "tsne",
        #     partial(sc.pl.tsne, color=["CD3D", "louvain"]),
        #     marks=pytest.mark.xfail(
        #         reason="slight differences even after setting random_state."
        #     ),
        # ),
        ("umap_nocolor", sc.pl.umap),
        (
            "umap",
            partial(
                sc.pl.umap,
                color=["louvain"],
                palette=["b", "grey80", "r", "yellow", "black", "gray", "lightblue"],
                frameon=False,
            ),
        ),
        (
            "umap_gene_expr",
            partial(
                sc.pl.umap,
                color=np.array(["LYZ", "CD79A"]),
                s=20,
                alpha=0.5,
                frameon=False,
                title=["gene1", "gene2"],
            ),
        ),
        (
            "umap_layer",
            partial(
                sc.pl.umap,
                color=np.array(["LYZ", "CD79A"]),
                s=20,
                alpha=0.5,
                frameon=False,
                title=["gene1", "gene2"],
                layer="test",
                vmin=100,
                vcenter=101,
            ),
        ),
        (
            "umap_with_edges",
            partial(sc.pl.umap, color="louvain", edges=True, edges_width=0.1, s=50),
        ),
        # ('diffmap', partial(sc.pl.diffmap, components='all', color=['CD3D'])),
        (
            "umap_symbols",
            partial(sc.pl.umap, color=["1", "2", "3"], gene_symbols="numbers"),
        ),
        (
            "pca_mask",
            partial(
                sc.pl.pca,
                color=["LYZ", "CD79A", "louvain"],
                mask_obs="mask",
            ),
        ),
    ],
)
def test_scatterplots(image_comparer, pbmc_scatterplots, id, fn):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    fn(pbmc_scatterplots, show=False)
    save_and_compare_images(id)


def test_scatter_embedding_groups_and_size(image_comparer):
    # test that the 'groups' parameter sorts
    # cells, such that the cells belonging to the groups are
    # plotted on top. This new ordering requires that the size
    # vector is also ordered (if given).
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.pl.embedding(
        pbmc,
        "umap",
        color=["bulk_labels"],
        groups=["CD14+ Monocyte", "Dendritic"],
        size=(np.arange(pbmc.shape[0]) / 40) ** 1.7,
    )
    save_and_compare_images("embedding_groups_size")


def test_scatter_embedding_add_outline_vmin_vmax_norm(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()

    sc.pl.embedding(
        pbmc,
        "X_umap",
        color=["percent_mito", "n_counts", "bulk_labels", "percent_mito"],
        s=200,
        frameon=False,
        add_outline=True,
        vmax=["p99.0", partial(np.percentile, q=90), None, 0.03],
        vmin=0.01,
        vcenter=[0.015, None, None, 0.025],
        outline_color=("#555555", "0.9"),
        outline_width=(0.5, 0.5),
        cmap="viridis_r",
        alpha=0.9,
        wspace=0.5,
    )
    save_and_compare_images("embedding_outline_vmin_vmax")


def test_scatter_embedding_add_outline_vmin_vmax_norm_ref(tmp_path, check_same_image):
    pbmc = pbmc68k_reduced()

    import matplotlib as mpl
    import matplotlib.pyplot as plt

    norm = mpl.colors.LogNorm()
    with pytest.raises(
        ValueError, match=r"Passing both norm and vmin/vmax/vcenter is not allowed\."
    ):
        sc.pl.embedding(
            pbmc,
            "X_umap",
            color=["percent_mito", "n_counts"],
            norm=norm,
            vmin=0,
            vmax=1,
            vcenter=0.5,
            cmap="RdBu_r",
        )

    try:
        from matplotlib.colors import TwoSlopeNorm as DivNorm
    except ImportError:
        # matplotlib<3.2
        from matplotlib.colors import DivergingNorm as DivNorm

    from matplotlib.colors import Normalize

    norm = Normalize(0, 10000)
    divnorm = DivNorm(200, 150, 6000)

    # allowed
    sc.pl.umap(
        pbmc,
        color=["n_counts", "bulk_labels", "percent_mito"],
        frameon=False,
        vmax=["p99.0", None, None],
        vcenter=[0.015, None, None],
        norm=[None, norm, norm],
        wspace=0.5,
    )

    sc.pl.umap(
        pbmc,
        color=["n_counts", "bulk_labels"],
        frameon=False,
        norm=norm,
        wspace=0.5,
    )
    plt.savefig(tmp_path / "umap_norm_fig0.png")
    plt.close()

    sc.pl.umap(
        pbmc,
        color=["n_counts", "bulk_labels"],
        frameon=False,
        norm=divnorm,
        wspace=0.5,
    )
    plt.savefig(tmp_path / "umap_norm_fig1.png")
    plt.close()

    sc.pl.umap(
        pbmc,
        color=["n_counts", "bulk_labels"],
        frameon=False,
        vcenter=200,
        vmin=150,
        vmax=6000,
        wspace=0.5,
    )
    plt.savefig(tmp_path / "umap_norm_fig2.png")
    plt.close()

    check_same_image(
        tmp_path / "umap_norm_fig1.png",
        tmp_path / "umap_norm_fig2.png",
        tol=1,
        root=tmp_path,
    )

    with pytest.raises(AssertionError):
        check_same_image(
            tmp_path / "umap_norm_fig1.png",
            tmp_path / "umap_norm_fig0.png",
            tol=1,
            root=tmp_path,
            save=False,
        )


def test_timeseries():
    adata = pbmc68k_reduced()
    sc.pp.neighbors(adata, n_neighbors=5, method="gauss", knn=False)
    sc.tl.diffmap(adata)
    sc.tl.dpt(adata, n_branchings=1, n_dcs=10)
    sc.pl.dpt_timeseries(adata, as_heatmap=True)


def test_scatter_raw(tmp_path):
    pbmc = pbmc68k_reduced()[:100].copy()
    raw_pth = tmp_path / "raw.png"
    x_pth = tmp_path / "X.png"

    sc.pl.scatter(pbmc, color="HES4", basis="umap", use_raw=True)
    plt.savefig(raw_pth, dpi=60)
    plt.close()

    sc.pl.scatter(pbmc, color="HES4", basis="umap", use_raw=False)
    plt.savefig(x_pth, dpi=60)
    plt.close()

    comp = compare_images(str(raw_pth), str(x_pth), tol=5)
    assert "Error" in comp, "Plots should change depending on use_raw."


def test_binary_scatter(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    data = AnnData(
        np.asarray([[-1, 2, 0], [3, 4, 0], [1, 2, 0]]).T,
        obs=dict(binary=np.asarray([False, True, True])),
    )
    sc.pp.pca(data)
    sc.pl.pca(data, color="binary")
    if pkg_version("scikit-learn") >= Version("1.5.0rc1"):
        save_and_compare_images("binary_pca")
    else:
        save_and_compare_images("binary_pca_old")


def test_scatter_specify_layer_and_raw():
    pbmc = pbmc68k_reduced()
    pbmc.layers["layer"] = pbmc.raw.X.copy()
    with pytest.raises(ValueError, match=r"Cannot use both a layer and.*raw"):
        sc.pl.umap(pbmc, color="HES4", use_raw=True, layer="layer")


@pytest.mark.parametrize(
    "color", ["n_genes", "bulk_labels", ["n_genes", "bulk_labels"]]
)
def test_scatter_no_basis_per_obs(image_comparer, color):
    """Test scatterplot of per-obs points with no basis."""
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.pl.scatter(
        pbmc,
        x="HES4",
        y="percent_mito",
        color=color,
        use_raw=False,
        # palette only applies to categorical, i.e. color=='bulk_labels'
        palette="Set2",
    )
    color_str = color if isinstance(color, str) else "_".join(color)
    save_and_compare_images(f"scatter_HES_percent_mito_{color_str}")


def test_scatter_no_basis_per_var(image_comparer):
    """Test scatterplot of per-var points with no basis."""
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.pl.scatter(pbmc, x="AAAGCCTGGCTAAC-1", y="AAATTCGATGCACA-1", use_raw=False)
    save_and_compare_images("scatter_AAAGCCTGGCTAAC-1_vs_AAATTCGATGCACA-1")


@pytest.fixture
def pbmc_filtered() -> Callable[[], AnnData]:
    pbmc = pbmc68k_reduced()
    sc.pp.filter_genes(pbmc, min_cells=10)
    return pbmc.copy


@pytest.mark.parametrize("use_raw", [True, None])
def test_scatter_no_basis_raw(check_same_image, pbmc_filtered, tmp_path, use_raw):
    """Test scatterplots of raw layer with no basis."""
    adata = pbmc_filtered()

    sc.pl.scatter(adata.raw.to_adata(), x="EGFL7", y="F12", color="FAM185A")
    plt.savefig(path1 := tmp_path / "scatter-raw-to-adata.png")

    sc.pl.scatter(adata, x="EGFL7", y="F12", color="FAM185A", use_raw=use_raw)
    plt.savefig(path2 := tmp_path / f"scatter-{use_raw=}.png")
    plt.close()

    check_same_image(path1, path2, tol=15, root=tmp_path)


@pytest.mark.parametrize(
    ("x", "y", "color", "use_raw"),
    [
        # test that plotting fails with a ValueError if trying to plot
        # var_names only found in raw and use_raw is False
        ("EGFL7", "F12", "FAM185A", False),
        # test that plotting fails if one axis is a per-var value and the
        # other is a per-obs value
        ("HES4", "n_cells", None, None),
        ("percent_mito", "AAAGCCTGGCTAAC-1", None, None),
    ],
)
def test_scatter_no_basis_value_error(pbmc_filtered, x, y, color, use_raw):
    """Test that `scatter()` raises `ValueError` where appropriate.

    If `sc.pl.scatter()` receives variable labels that either cannot be
    found or are incompatible with one another, the function should
    raise a `ValueError`. This test checks that this happens as
    expected.
    """
    with pytest.raises(
        ValueError, match=r"inputs must all come from either `\.obs` or `\.var`"
    ):
        sc.pl.scatter(pbmc_filtered(), x=x, y=y, color=color, use_raw=use_raw)


def test_rankings(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()
    sc.pp.pca(pbmc)
    sc.pl.pca_loadings(pbmc)
    save_and_compare_images("pca_loadings")

    sc.pl.pca_loadings(pbmc, components="1,2,3")
    save_and_compare_images("pca_loadings")

    sc.pl.pca_loadings(pbmc, components=[1, 2, 3])
    save_and_compare_images("pca_loadings")

    sc.pl.pca_loadings(pbmc, include_lowest=False)
    save_and_compare_images("pca_loadings_without_lowest")

    sc.pl.pca_loadings(pbmc, n_points=10)
    save_and_compare_images("pca_loadings_10_points")


# TODO: Make more generic
def test_scatter_rep(tmp_path):
    """Test to make sure I can predict when scatter reps should be the same."""
    rep_args = {
        "raw": {"use_raw": True},
        "layer": {"layer": "layer", "use_raw": False},
        "X": {"use_raw": False},
    }
    states = pd.DataFrame.from_records(
        zip(
            list(chain.from_iterable(repeat(x, 3) for x in ["X", "raw", "layer"])),
            list(chain.from_iterable(repeat("abc", 3))),
            [1, 2, 3, 3, 1, 2, 2, 3, 1],
            strict=True,
        ),
        columns=["rep", "gene", "result"],
    )
    states["outpth"] = [
        tmp_path / f"{state.gene}_{state.rep}_{state.result}.png"
        for state in states.itertuples()
    ]
    pattern = np.array(list(chain.from_iterable(repeat(i, 5) for i in range(3))))
    coords = np.c_[np.arange(15) % 5, pattern]

    adata = AnnData(
        X=np.zeros((15, 3)),
        layers={"layer": np.zeros((15, 3))},
        obsm={"X_pca": coords},
        var=pd.DataFrame(index=list("abc")),
        obs=pd.DataFrame(index=[f"cell{i}" for i in range(15)]),
    )
    adata.raw = adata.copy()
    adata.X[np.arange(15), pattern] = 1
    adata.raw.X[np.arange(15), (pattern + 1) % 3] = 1
    adata.layers["layer"][np.arange(15), (pattern + 2) % 3] = 1

    for state in states.itertuples():
        sc.pl.pca(adata, color=state.gene, **rep_args[state.rep], show=False)
        plt.savefig(state.outpth, dpi=60)
        plt.close()

    for s1, s2 in combinations(states.itertuples(), 2):
        comp = compare_images(str(s1.outpth), str(s2.outpth), tol=5)
        if s1.result == s2.result:
            assert comp is None, comp
        else:
            assert "Error" in comp, f"{s1.outpth}, {s2.outpth} aren't supposed to match"


def test_no_copy():
    # https://github.com/scverse/scanpy/issues/1000
    # Tests that plotting functions don't make a copy from a view unless they
    # actually have to
    actual = pbmc68k_reduced()
    sc.pl.umap(actual, color=["bulk_labels", "louvain"], show=False)  # Set colors

    view = actual[np.random.choice(actual.obs_names, size=actual.shape[0] // 5), :]

    sc.pl.umap(view, color=["bulk_labels", "louvain"], show=False)
    assert view.is_view

    rank_genes_groups_plotting_funcs = [
        sc.pl.rank_genes_groups,
        sc.pl.rank_genes_groups_dotplot,
        sc.pl.rank_genes_groups_heatmap,
        sc.pl.rank_genes_groups_matrixplot,
        sc.pl.rank_genes_groups_stacked_violin,
        # TODO: raises ValueError about empty distance matrix – investigate
        # sc.pl.rank_genes_groups_tracksplot,
        sc.pl.rank_genes_groups_violin,
    ]

    # the pbmc68k was generated using rank_genes_groups with method='logreg'
    # which does not generate 'logfoldchanges', although this field is
    # required by `sc.get.rank_genes_groups_df`.
    # After updating rank_genes_groups plots to use the latter function
    # an error appears. Re-running rank_genes_groups with default method
    # solves the problem.
    sc.tl.rank_genes_groups(actual, "bulk_labels")

    # Only plotting one group at a time to avoid generating dendrogram
    # TODO: Generating a dendrogram modifies the object, this should be
    # optional and also maybe not modify the object.
    for plotfunc in rank_genes_groups_plotting_funcs:
        view = actual[actual.obs["bulk_labels"] == "Dendritic"]
        plotfunc(view, ["Dendritic"], show=False)
        assert view.is_view


def test_groupby_index(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=15)

    pbmc = pbmc68k_reduced()

    genes = [
        "CD79A",
        "MS4A1",
        "CD8A",
        "CD8B",
        "LYZ",
        "LGALS3",
        "S100A8",
        "GNLY",
        "NKG7",
        "KLRB1",
        "FCGR3A",
        "FCER1A",
        "CST3",
    ]
    pbmc_subset = pbmc[:10].copy()
    sc.pl.dotplot(pbmc_subset, genes, groupby="index")
    save_and_compare_images("dotplot_groupby_index")


# test category order when groupby is a list (#1735)
def test_groupby_list(image_comparer):
    save_and_compare_images = partial(image_comparer, ROOT, tol=30)

    adata = krumsiek11()

    np.random.seed(1)

    cat_val = adata.obs.cell_type.tolist()
    np.random.shuffle(cat_val)
    cats = adata.obs.cell_type.cat.categories.tolist()
    np.random.shuffle(cats)
    adata.obs["rand_cat"] = pd.Categorical(cat_val, categories=cats)

    with mpl.rc_context({"figure.subplot.bottom": 0.5}):
        sc.pl.dotplot(
            adata, ["Gata1", "Gata2"], groupby=["rand_cat", "cell_type"], swap_axes=True
        )
        save_and_compare_images("dotplot_groupby_list_catorder")


def test_color_cycler(caplog):
    # https://github.com/scverse/scanpy/issues/1885
    import logging

    pbmc = pbmc68k_reduced()
    colors = sns.color_palette("deep")
    cyl = sns.rcmod.cycler("color", sns.color_palette("deep"))

    with (
        caplog.at_level(logging.WARNING),
        plt.rc_context({"axes.prop_cycle": cyl, "patch.facecolor": colors[0]}),
    ):
        sc.pl.umap(pbmc, color="phase")
        plt.show()
        plt.close()

    assert caplog.text == ""


def test_repeated_colors_w_missing_value():
    # https://github.com/scverse/scanpy/issues/2133
    v = pd.Series(np.arange(10).astype(str))
    v[0] = np.nan
    v = v.astype("category")

    ad = sc.AnnData(obs=pd.DataFrame(v, columns=["value"]))
    ad.obsm["X_umap"] = np.random.normal(size=(ad.n_obs, 2))

    sc.pl.umap(ad, color="value")

    ad.uns["value_colors"][1] = ad.uns["value_colors"][0]

    sc.pl.umap(ad, color="value")


@pytest.mark.parametrize(
    "plot",
    [
        sc.pl.rank_genes_groups_dotplot,
        sc.pl.rank_genes_groups_heatmap,
        sc.pl.rank_genes_groups_matrixplot,
        sc.pl.rank_genes_groups_stacked_violin,
        sc.pl.rank_genes_groups_tracksplot,
        # TODO: add other rank_genes_groups plots here once they work
    ],
)
def test_filter_rank_genes_groups_plots(tmp_path, plot, check_same_image):
    n_genes = 4

    adata = pbmc68k_reduced()

    sc.tl.rank_genes_groups(adata, "bulk_labels", method="wilcoxon", pts=True)

    sc.tl.filter_rank_genes_groups(
        adata,
        key_added="rank_genes_groups_filtered",
        min_in_group_fraction=0.25,
        min_fold_change=1,
        max_out_group_fraction=0.5,
    )

    conditions = "logfoldchanges >= 1 & pct_nz_group >= .25 & pct_nz_reference < .5"
    df = sc.get.rank_genes_groups_df(adata, group=None, key="rank_genes_groups")
    df = df.query(conditions)[["group", "names"]]

    var_names = {
        k: v.head(n_genes).tolist()
        for k, v in df.groupby("group", observed=True)["names"]
    }

    pth_a = tmp_path / f"{plot.__name__}_filter_a.png"
    pth_b = tmp_path / f"{plot.__name__}_filter_b.png"

    plot(adata, key="rank_genes_groups_filtered", n_genes=n_genes)
    plt.savefig(pth_a)
    plt.close()

    plot(adata, key="rank_genes_groups", var_names=var_names)
    plt.savefig(pth_b)
    plt.close()

    check_same_image(pth_a, pth_b, tol=1, root=tmp_path)


@needs.skmisc
@pytest.mark.parametrize(
    ("id", "params"),
    [
        pytest.param("scrublet", {}, id="scrublet"),
        pytest.param("scrublet_no_threshold", {}, id="scrublet_no_threshold"),
        pytest.param(
            "scrublet_with_batches", dict(batch_key="batch"), id="scrublet_with_batches"
        ),
    ],
)
def test_scrublet_plots(monkeypatch, image_comparer, id, params):
    save_and_compare_images = partial(image_comparer, ROOT, tol=10)

    adata = pbmc3k()[:200].copy()
    adata.obs["batch"] = 100 * ["a"] + 100 * ["b"]

    with monkeypatch.context() as m:
        if id == "scrublet_no_threshold":
            m.setattr("skimage.filters.threshold_minimum", None)
        sc.pp.scrublet(adata, use_approx_neighbors=False, **params)
    if id == "scrublet_no_threshold":
        assert "threshold" not in adata.uns["scrublet"]

    sc.pl.scrublet_score_distribution(adata, return_fig=True, show=False)
    save_and_compare_images(id)


def test_umap_mask_equal(tmp_path, check_same_image):
    """Check that all desired cells are coloured and masked cells gray."""
    pbmc = pbmc3k_processed()
    mask_obs = pbmc.obs["louvain"].isin(["B cells", "NK cells"])

    ax = sc.pl.umap(pbmc, size=8.0, show=False)
    sc.pl.umap(pbmc[mask_obs], size=8.0, color="LDHB", ax=ax)
    plt.savefig(p1 := tmp_path / "umap_mask_fig1.png")
    plt.close()

    sc.pl.umap(pbmc, size=8.0, color="LDHB", mask_obs=mask_obs)
    plt.savefig(p2 := tmp_path / "umap_mask_fig2.png")
    plt.close()

    check_same_image(p1, p2, tol=1, root=tmp_path)


def test_umap_mask_mult_plots():
    """Check that multiple images are plotted when color is a list."""
    pbmc = pbmc3k_processed()
    color = ["LDHB", "LYZ", "CD79A"]
    mask_obs = pbmc.obs["louvain"].isin(["B cells", "NK cells"])
    axes = sc.pl.umap(pbmc, color=color, mask_obs=mask_obs, show=False)
    assert isinstance(axes, list)
    assert len(axes) == len(color)


def test_umap_categories_dont_change_when_rerun_with_fewer_categories():
    """Check that lowering the categories of interest does not cause a recalculation of colors."""
    pbmc = pbmc3k_processed()
    _ = sc.pl.umap(pbmc, color="louvain", show=False)
    assert len(pbmc.uns["louvain_colors"]) == len(pbmc.obs["louvain"].cat.categories)
    old_colors = pbmc.uns["louvain_colors"].copy()
    pbmc.obs.loc[pbmc.obs["louvain"] == "NK cells", "louvain"] = "B cells"
    pbmc.obs["louvain"] = pbmc.obs["louvain"].cat.remove_unused_categories()
    # see https://github.com/scverse/scanpy/issues/3716 for why this used to fail
    # Recalculation of the UMAP should not cause a re-calculation of colors
    # when there are fewer categories.
    _ = sc.pl.umap(pbmc, color="louvain", show=False)
    assert (old_colors == pbmc.uns["louvain_colors"]).all()


def test_umap_categories_change_when_rerun_with_more_categories():
    """Check that growing the categories of interest causes a recalculation of colors."""
    pbmc = pbmc3k_processed()
    _ = sc.pl.umap(pbmc, color="louvain", show=False)
    assert len(pbmc.uns["louvain_colors"]) == len(pbmc.obs["louvain"].cat.categories)
    pbmc.obs["louvain"] = pbmc.obs["louvain"].cat.add_categories("New Category")
    pbmc.obs.loc[pbmc.obs_names[:5], "louvain"] = "New Category"
    _ = sc.pl.umap(pbmc, color="louvain", show=False)
    assert len(pbmc.obs["louvain"].cat.categories) == len(pbmc.uns["louvain_colors"])


def test_umap_mask_no_modification():
    """Check that mask_obs argument doesn't affect the data being plotted."""
    pbmc = pbmc3k_processed()
    data_copy = pbmc.obs["louvain"].copy()
    sc.pl.umap(
        pbmc, mask_obs=(pbmc.obs["louvain"] == "B cells"), color="louvain", show=False
    )
    pd.testing.assert_series_equal(pbmc.obs["louvain"], data_copy)


def test_string_mask(tmp_path, check_same_image):
    """Check that the same mask given as string or bool array provides the same result."""
    pbmc = pbmc3k_processed()
    pbmc.obs["mask"] = mask_obs = pbmc.obs["louvain"].isin(["B cells", "NK cells"])

    sc.pl.umap(pbmc, mask_obs=mask_obs, color="LDHB")
    plt.savefig(p1 := tmp_path / "umap_mask_fig1.png")
    plt.close()

    sc.pl.umap(pbmc, color="LDHB", mask_obs="mask")
    plt.savefig(p2 := tmp_path / "umap_mask_fig2.png")
    plt.close()

    check_same_image(p1, p2, tol=1, root=tmp_path)


def test_violin_scale_warning(monkeypatch):
    adata = pbmc3k_processed()
    monkeypatch.setattr(sc.pl.StackedViolin, "DEFAULT_SCALE", "count", raising=False)
    with pytest.warns(FutureWarning, match="Don’t set DEFAULT_SCALE"):
        sc.pl.StackedViolin(adata, adata.var_names[:3], groupby="louvain")



"""Biomart queries."""

from __future__ import annotations

from ._queries import (
    biomart_annotations,
    enrich,  # gprofiler queries
    gene_coordinates,
    mitochondrial_genes,
)

__all__ = [
    "biomart_annotations",
    "enrich",
    "gene_coordinates",
    "mitochondrial_genes",
]



from __future__ import annotations

import anndata as ad
import numpy as np
import pandas as pd
import pytest
from packaging.version import Version
from scipy import sparse

import scanpy as sc
from scanpy._compat import pkg_version
from scanpy._utils import _resolve_axis, get_literal_vals
from scanpy.get._aggregated import AggType
from testing.scanpy._helpers import assert_equal
from testing.scanpy._helpers.data import pbmc3k_processed
from testing.scanpy._pytest.params import ARRAY_TYPES_MEM


@pytest.fixture(params=get_literal_vals(AggType))
def metric(request: pytest.FixtureRequest) -> AggType:
    return request.param


@pytest.fixture
def df_base():
    ax_base = ["A", "B"]
    return pd.DataFrame(index=ax_base)


@pytest.fixture
def df_groupby():
    ax_groupby = [
        *["v0", "v1", "v2"],
        *["w0", "w1"],
        *["a1", "a2", "a3"],
        *["b1", "b2"],
        *["c1", "c2"],
        "d0",
    ]

    df_groupby = pd.DataFrame(index=pd.Index(ax_groupby, name="cell"))
    df_groupby["key"] = pd.Categorical([c[0] for c in ax_groupby])
    df_groupby["key_superset"] = pd.Categorical([c[0] for c in ax_groupby]).map({
        **{"v": "v", "w": "v"},  # noqa: PIE800
        **{"a": "a", "b": "a", "c": "a", "d": "a"},  # noqa: PIE800
    })
    df_groupby["key_subset"] = pd.Categorical([c[1] for c in ax_groupby])
    df_groupby["weight"] = 2.0
    return df_groupby


@pytest.fixture
def x():
    data = [
        *[[0, -2], [1, 13], [2, 1]],  # v
        *[[3, 12], [4, 2]],  # w
        *[[5, 11], [6, 3], [7, 10]],  # a
        *[[8, 4], [9, 9]],  # b
        *[[10, 5], [11, 8]],  # c
        [12, 6],  # d
    ]
    return np.array(data, dtype=np.float32)


def gen_adata(data_key, dim, df_base, df_groupby, x):
    if (data_key == "varm" and dim == "obs") or (data_key == "obsm" and dim == "var"):
        pytest.skip("invalid parameter combination")

    obs_df, var_df = (df_groupby, df_base) if dim == "obs" else (df_base, df_groupby)
    data = x.T if dim == "var" and data_key != "varm" else x
    if data_key != "X":
        data_dict_sparse = {data_key: {"test": sparse.csr_matrix(data)}}  # noqa: TID251
        data_dict_dense = {data_key: {"test": data}}
    else:
        data_dict_sparse = {data_key: sparse.csr_matrix(data)}  # noqa: TID251
        data_dict_dense = {data_key: data}

    adata_sparse = ad.AnnData(obs=obs_df, var=var_df, **data_dict_sparse)
    adata_dense = ad.AnnData(obs=obs_df, var=var_df, **data_dict_dense)
    return adata_sparse, adata_dense


@pytest.mark.parametrize("axis", [0, 1])
def test_mask(axis):
    blobs = sc.datasets.blobs()
    mask = blobs.obs["blobs"] == 0
    blobs.obs["mask_col"] = mask
    if axis == 1:
        blobs = blobs.T
    by_name = sc.get.aggregate(blobs, "blobs", "sum", axis=axis, mask="mask_col")
    by_value = sc.get.aggregate(blobs, "blobs", "sum", axis=axis, mask=mask)

    assert_equal(by_name, by_value)

    assert np.all(by_name["0"].layers["sum"] == 0)


@pytest.mark.parametrize("array_type", ARRAY_TYPES_MEM)
def test_aggregate_vs_pandas(metric, array_type):
    adata = pbmc3k_processed().raw.to_adata()
    adata = adata[
        adata.obs["louvain"].isin(adata.obs["louvain"].cat.categories[:5]), :1_000
    ].copy()
    adata.X = array_type(adata.X)
    adata.obs["percent_mito_binned"] = pd.cut(adata.obs["percent_mito"], bins=5)
    result = sc.get.aggregate(adata, ["louvain", "percent_mito_binned"], metric)

    if metric == "count_nonzero":
        expected = (
            (adata.to_df() != 0)
            .astype(np.float64)
            .join(adata.obs[["louvain", "percent_mito_binned"]])
            .groupby(["louvain", "percent_mito_binned"], observed=True)
            .agg("sum")
        )
    else:
        expected = (
            adata.to_df()
            .astype(np.float64)
            .join(adata.obs[["louvain", "percent_mito_binned"]])
            .groupby(["louvain", "percent_mito_binned"], observed=True)
            .agg(metric)
        )
    expected.index = expected.index.to_frame().apply(
        lambda x: "_".join(map(str, x)), axis=1
    )
    expected.index.name = None
    expected.columns.name = None

    result_df = result.to_df(layer=metric)
    result_df.index.name = None
    result_df.columns.name = None

    if pkg_version("pandas") < Version("2"):
        # Order of results returned by groupby changed in pandas 2
        assert expected.shape == result_df.shape
        assert expected.index.isin(result_df.index).all()

        expected = expected.loc[result_df.index]

    pd.testing.assert_frame_equal(result_df, expected, check_dtype=False, atol=1e-5)


@pytest.mark.parametrize("array_type", ARRAY_TYPES_MEM)
def test_aggregate_axis(array_type, metric):
    adata = pbmc3k_processed().raw.to_adata()
    adata = adata[
        adata.obs["louvain"].isin(adata.obs["louvain"].cat.categories[:5]), :1_000
    ].copy()
    adata.X = array_type(adata.X)
    expected = sc.get.aggregate(adata, ["louvain"], metric)
    actual = sc.get.aggregate(adata.T, ["louvain"], metric, axis=1).T

    assert_equal(expected, actual)


def test_aggregate_entry():
    args = ("blobs", ["mean", "var", "count_nonzero"])

    adata = sc.datasets.blobs()
    x_result = sc.get.aggregate(adata, *args)
    # layer adata
    layer_adata = ad.AnnData(
        obs=adata.obs,
        var=adata.var,
        layers={"test": adata.X.copy()},
    )
    layer_result = sc.get.aggregate(layer_adata, *args, layer="test")
    obsm_adata = ad.AnnData(
        obs=adata.obs,
        var=adata.var,
        obsm={"test": adata.X.copy()},
    )
    obsm_result = sc.get.aggregate(obsm_adata, *args, obsm="test")
    varm_adata = ad.AnnData(
        obs=adata.var,
        var=adata.obs,
        varm={"test": adata.X.copy()},
    )
    varm_result = sc.get.aggregate(varm_adata, *args, varm="test")

    x_result_min = x_result.copy()
    del x_result_min.var
    x_result_min.var_names = [str(x) for x in np.arange(x_result_min.n_vars)]

    assert_equal(x_result, layer_result)
    assert_equal(x_result_min, obsm_result)
    assert_equal(x_result.layers, obsm_result.layers)
    assert_equal(x_result.layers, varm_result.T.layers)


def test_aggregate_incorrect_dim():
    adata = pbmc3k_processed().raw.to_adata()

    with pytest.raises(ValueError, match="was 'foo'"):
        sc.get.aggregate(adata, ["louvain"], "sum", axis="foo")


@pytest.mark.parametrize("axis_name", ["obs", "var"])
def test_aggregate_axis_specification(axis_name):
    axis, axis_name = _resolve_axis(axis_name)
    by = "blobs" if axis == 0 else "labels"

    adata = sc.datasets.blobs()
    adata.var["labels"] = np.tile(["a", "b"], adata.shape[1])[: adata.shape[1]]

    agg_index = sc.get.aggregate(adata, by=by, func="mean", axis=axis)
    agg_name = sc.get.aggregate(adata, by=by, func="mean", axis=axis_name)

    np.testing.assert_equal(agg_index.layers["mean"], agg_name.layers["mean"])

    if axis_name == "obs":
        agg_unspecified = sc.get.aggregate(adata, by=by, func="mean")
        np.testing.assert_equal(agg_name.layers["mean"], agg_unspecified.layers["mean"])


@pytest.mark.parametrize(
    ("matrix", "df", "keys", "metrics", "expected"),
    [
        pytest.param(
            np.block([
                [np.ones((2, 2)), np.zeros((2, 2))],
                [np.zeros((2, 2)), np.ones((2, 2))],
            ]),
            pd.DataFrame(
                {
                    "a": ["a", "a", "b", "b"],
                    "b": ["c", "d", "d", "d"],
                },
                index=["a_c", "a_d", "b_d1", "b_d2"],
            ),
            ["a", "b"],
            ["count_nonzero"],  # , "sum", "mean"],
            ad.AnnData(
                obs=pd.DataFrame(
                    {"a": ["a", "a", "b"], "b": ["c", "d", "d"]},
                    index=["a_c", "a_d", "b_d"],
                ).astype("category"),
                var=pd.DataFrame(index=[f"gene_{i}" for i in range(4)]),
                layers={
                    "count_nonzero": np.array([
                        [1, 1, 0, 0],
                        [1, 1, 0, 0],
                        [0, 0, 2, 2],
                    ]),
                    # "sum": np.array([[2, 0], [0, 2]]),
                    # "mean": np.array([[1, 0], [0, 1]]),
                },
            ),
            id="count_nonzero",
        ),
        pytest.param(
            np.block([
                [np.ones((2, 2)), np.zeros((2, 2))],
                [np.zeros((2, 2)), np.ones((2, 2))],
            ]),
            pd.DataFrame(
                {
                    "a": ["a", "a", "b", "b"],
                    "b": ["c", "d", "d", "d"],
                },
                index=["a_c", "a_d", "b_d1", "b_d2"],
            ),
            ["a", "b"],
            ["sum", "mean", "count_nonzero"],
            ad.AnnData(
                obs=pd.DataFrame(
                    {"a": ["a", "a", "b"], "b": ["c", "d", "d"]},
                    index=["a_c", "a_d", "b_d"],
                ).astype("category"),
                var=pd.DataFrame(index=[f"gene_{i}" for i in range(4)]),
                layers={
                    "sum": np.array([[1, 1, 0, 0], [1, 1, 0, 0], [0, 0, 2, 2]]),
                    "mean": np.array([[1, 1, 0, 0], [1, 1, 0, 0], [0, 0, 1, 1]]),
                    "count_nonzero": np.array([
                        [1, 1, 0, 0],
                        [1, 1, 0, 0],
                        [0, 0, 2, 2],
                    ]),
                },
            ),
            id="sum-mean-count_nonzero",
        ),
        pytest.param(
            np.block([
                [np.ones((2, 2)), np.zeros((2, 2))],
                [np.zeros((2, 2)), np.ones((2, 2))],
            ]),
            pd.DataFrame(
                {
                    "a": ["a", "a", "b", "b"],
                    "b": ["c", "d", "d", "d"],
                },
                index=["a_c", "a_d", "b_d1", "b_d2"],
            ),
            ["a", "b"],
            ["mean"],
            ad.AnnData(
                obs=pd.DataFrame(
                    {"a": ["a", "a", "b"], "b": ["c", "d", "d"]},
                    index=["a_c", "a_d", "b_d"],
                ).astype("category"),
                var=pd.DataFrame(index=[f"gene_{i}" for i in range(4)]),
                layers={
                    "mean": np.array([[1, 1, 0, 0], [1, 1, 0, 0], [0, 0, 1, 1]]),
                },
            ),
            id="mean",
        ),
    ],
)
def test_aggregate_examples(matrix, df, keys, metrics, expected):
    adata = ad.AnnData(
        X=matrix,
        obs=df,
        var=pd.DataFrame(index=[f"gene_{i}" for i in range(matrix.shape[1])]),
    )
    result = sc.get.aggregate(adata, by=keys, func=metrics)

    print(result)
    print(expected)

    assert_equal(expected, result)


@pytest.mark.parametrize(
    ("label_cols", "cols", "expected"),
    [
        pytest.param(
            dict(
                a=pd.Categorical(["a", "b", "c"]),
                b=pd.Categorical(["d", "d", "f"]),
            ),
            ["a", "b"],
            pd.Categorical(["a_d", "b_d", "c_f"]),
            id="two_of_two",
        ),
        pytest.param(
            dict(
                a=pd.Categorical(["a", "b", "c"]),
                b=pd.Categorical(["d", "d", "f"]),
                c=pd.Categorical(["g", "h", "h"]),
            ),
            ["a", "b", "c"],
            pd.Categorical(["a_d_g", "b_d_h", "c_f_h"]),
            id="three_of_three",
        ),
        pytest.param(
            dict(
                a=pd.Categorical(["a", "b", "c"]),
                b=pd.Categorical(["d", "d", "f"]),
                c=pd.Categorical(["g", "h", "h"]),
            ),
            ["a", "c"],
            pd.Categorical(["a_g", "b_h", "c_h"]),
            id="two_of_three-1",
        ),
        pytest.param(
            dict(
                a=pd.Categorical(["a", "b", "c"]),
                b=pd.Categorical(["d", "d", "f"]),
                c=pd.Categorical(["g", "h", "h"]),
            ),
            ["b", "c"],
            pd.Categorical(["d_g", "d_h", "f_h"]),
            id="two_of_three-2",
        ),
    ],
)
def test_combine_categories(label_cols, cols, expected):
    from scanpy.get._aggregated import _combine_categories

    label_df = pd.DataFrame(label_cols)
    result, result_label_df = _combine_categories(label_df, cols)

    assert isinstance(result, pd.Categorical)

    pd.testing.assert_extension_array_equal(result, expected)

    pd.testing.assert_index_equal(
        pd.Index(result), result_label_df.index.astype("category")
    )

    reconstructed_df = pd.DataFrame(
        [x.split("_") for x in result], columns=cols, index=result.astype(str)
    ).astype("category")
    pd.testing.assert_frame_equal(reconstructed_df, result_label_df)


@pytest.mark.parametrize("array_type", ARRAY_TYPES_MEM)
def test_aggregate_arraytype(array_type, metric):
    adata = pbmc3k_processed().raw.to_adata()
    adata = adata[
        adata.obs["louvain"].isin(adata.obs["louvain"].cat.categories[:5]), :1_000
    ].copy()
    adata.X = array_type(adata.X)
    aggregate = sc.get.aggregate(adata, ["louvain"], metric)
    assert isinstance(aggregate.layers[metric], np.ndarray)


def test_aggregate_obsm_varm():
    adata_obsm = sc.datasets.blobs()
    adata_obsm.obs["blobs"] = adata_obsm.obs["blobs"].astype(str)
    adata_obsm.obsm["test"] = adata_obsm.X[:, ::2].copy()
    adata_varm = adata_obsm.T.copy()

    result_obsm = sc.get.aggregate(adata_obsm, "blobs", ["sum", "mean"], obsm="test")
    result_varm = sc.get.aggregate(adata_varm, "blobs", ["sum", "mean"], varm="test")

    assert_equal(result_obsm, result_varm.T)

    expected_sum = (
        pd.DataFrame(adata_obsm.obsm["test"], index=adata_obsm.obs_names)
        .groupby(adata_obsm.obs["blobs"], observed=True)
        .sum()
    )
    expected_mean = (
        pd.DataFrame(adata_obsm.obsm["test"], index=adata_obsm.obs_names)
        .groupby(adata_obsm.obs["blobs"], observed=True)
        .mean()
    )

    assert_equal(expected_sum.values, result_obsm.layers["sum"])
    assert_equal(expected_mean.values, result_obsm.layers["mean"])


def test_aggregate_obsm_labels():
    from itertools import chain, repeat

    label_counts = [("a", 5), ("b", 3), ("c", 4)]
    blocks = [np.ones((n, 1)) for _, n in label_counts]
    obs_names = pd.Index([
        f"cell_{i:02d}" for i in range(sum(b.shape[0] for b in blocks))
    ])
    entry = pd.DataFrame(
        sparse.block_diag(blocks).toarray(),
        columns=[f"dim_{i}" for i in range(len(label_counts))],
        index=obs_names,
    )

    adata = ad.AnnData(
        obs=pd.DataFrame(
            {
                "labels": list(
                    chain.from_iterable(repeat(l, n) for (l, n) in label_counts)
                )
            },
            index=obs_names,
        ),
        var=pd.DataFrame(index=["gene_0"]),
        obsm={"entry": entry},
    )

    expected = ad.AnnData(
        obs=pd.DataFrame({"labels": pd.Categorical(list("abc"))}, index=list("abc")),
        var=pd.DataFrame(index=[f"dim_{i}" for i in range(3)]),
        layers={
            "sum": np.diag([n for _, n in label_counts]),
        },
    )
    result = sc.get.aggregate(adata, by="labels", func="sum", obsm="entry")
    assert_equal(expected, result)


def test_dispatch_not_implemented():
    adata = sc.datasets.blobs()
    with pytest.raises(NotImplementedError):
        sc.get.aggregate(adata.X, adata.obs["blobs"], "sum")


def test_factors():
    from itertools import product

    obs = pd.DataFrame(product(range(5), repeat=4), columns=list("abcd"))
    obs.index = [f"cell_{i:04d}" for i in range(obs.shape[0])]
    adata = ad.AnnData(
        X=np.arange(obs.shape[0]).reshape(-1, 1),
        obs=obs,
    )

    res = sc.get.aggregate(adata, by=["a", "b", "c", "d"], func="sum")
    np.testing.assert_equal(res.layers["sum"], adata.X)



from __future__ import annotations

import numpy as np
import pandas as pd
from anndata import AnnData

import scanpy.external as sce


def test_cell_demultiplexing():
    import random

    from scipy import stats

    random.seed(52)
    signal = stats.poisson.rvs(1000, 1, 990)
    doublet_signal = stats.poisson.rvs(1000, 1, 10)
    x = np.reshape(stats.poisson.rvs(500, 1, 10000), (1000, 10))
    for idx, signal_count in enumerate(signal):
        col_pos = idx % 10
        x[idx, col_pos] = signal_count

    for idx, signal_count in enumerate(doublet_signal):
        col_pos = (idx % 10) - 1
        x[idx, col_pos] = signal_count

    test_data = AnnData(np.random.randint(0, 100, size=x.shape), obs=pd.DataFrame(x))
    sce.pp.hashsolo(test_data, test_data.obs.columns)

    doublets = ["Doublet"] * 10
    classes = np.repeat(np.arange(10), 98).reshape(98, 10, order="F").ravel().tolist()
    negatives = ["Negative"] * 10
    expected = pd.array(doublets + classes + negatives, dtype="string")
    classification = test_data.obs["Classification"].array.astype("string")
    # This is a bit flaky, so allow some mismatches:
    if (expected != classification).sum() > 3:
        # Compare lists for better error message
        assert classification.tolist() == expected.tolist()



from __future__ import annotations

from contextlib import contextmanager
from enum import EnumMeta, IntEnum
from logging import getLevelName
from typing import TYPE_CHECKING

from .._compat import deprecated

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Literal

    _VerbosityName = Literal["error", "warning", "info", "hint", "debug"]
    _LoggingLevelName = Literal["CRITICAL", "ERROR", "WARNING", "INFO", "HINT", "DEBUG"]


_VERBOSITY_TO_LOGLEVEL: dict[int | _VerbosityName, _LoggingLevelName] = {
    "error": "ERROR",
    "warning": "WARNING",
    "info": "INFO",
    "hint": "HINT",
    "debug": "DEBUG",
}
_VERBOSITY_TO_LOGLEVEL.update(dict(enumerate(list(_VERBOSITY_TO_LOGLEVEL.values()))))


class VerbosityMeta(EnumMeta):
    @property
    @deprecated("Use `Verbosity.warning` instead")
    def warn(cls) -> Verbosity:
        return Verbosity.warning


class Verbosity(IntEnum, metaclass=VerbosityMeta):
    """Logging verbosity levels for :attr:`scanpy.settings.verbosity`."""

    error = 0
    """Error (`0`)"""
    warning = 1
    """Warning (`1`)"""
    info = 2
    """Info (`2`)"""
    hint = 3
    """Hint (`3`)"""
    debug = 4
    """Debug (`4`)"""

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Verbosity):
            return self is other
        if isinstance(other, int):
            return self.value == other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __hash__(self) -> int:
        # See https://docs.astral.sh/ruff/rules/eq-without-hash/
        return super().__hash__()

    @property
    def level(self) -> int:
        """The :ref:`logging level <levels>` corresponding to this verbosity level."""
        # getLevelName(str) returns the int level…
        return getLevelName(_VERBOSITY_TO_LOGLEVEL[self.name])

    @contextmanager
    def override(
        self, verbosity: Verbosity | _VerbosityName | int
    ) -> Generator[Verbosity, None, None]:
        """Temporarily override verbosity.

        >>> import scanpy as sc
        >>> sc.settings.verbosity = sc.Verbosity.info
        >>> with sc.settings.verbosity.override(sc.settings.verbosity.debug):
        ...     sc.settings.verbosity
        <Verbosity.debug: 4>
        >>> sc.settings.verbosity
        <Verbosity.info: 2>
        """
        from scanpy import settings

        settings.verbosity = verbosity
        try:
            yield self
        finally:
            settings.verbosity = self



"""Use harmony to integrate cells from different experiments."""

from __future__ import annotations

from collections.abc import Sequence  # noqa: TC003
from typing import TYPE_CHECKING

import numpy as np

from ..._compat import old_positionals
from ..._utils._doctests import doctest_needs

if TYPE_CHECKING:
    from anndata import AnnData


@old_positionals("basis", "adjusted_basis")
@doctest_needs("harmonypy")
def harmony_integrate(
    adata: AnnData,
    key: str | Sequence[str],
    *,
    basis: str = "X_pca",
    adjusted_basis: str = "X_pca_harmony",
    **kwargs,
):
    """Use harmonypy :cite:p:`Korsunsky2019` to integrate different experiments.

    Harmony :cite:p:`Korsunsky2019` is an algorithm for integrating single-cell
    data from multiple experiments. This function uses the python
    port of Harmony, ``harmonypy``, to integrate single-cell data
    stored in an AnnData object. As Harmony works by adjusting the
    principal components, this function should be run after performing
    PCA but before computing the neighbor graph, as illustrated in the
    example below.

    Parameters
    ----------
    adata
        The annotated data matrix.
    key
        The name of the column in ``adata.obs`` that differentiates
        among experiments/batches. To integrate over two or more covariates,
        you can pass multiple column names as a list. See ``vars_use``
        parameter of the ``harmonypy`` package for more details.
    basis
        The name of the field in ``adata.obsm`` where the PCA table is
        stored. Defaults to ``'X_pca'``, which is the default for
        ``sc.pp.pca()``.
    adjusted_basis
        The name of the field in ``adata.obsm`` where the adjusted PCA
        table will be stored after running this function. Defaults to
        ``X_pca_harmony``.
    kwargs
        Any additional arguments will be passed to
        ``harmonypy.run_harmony()``.

    Returns
    -------
    Updates adata with the field ``adata.obsm[obsm_out_field]``,
    containing principal components adjusted by Harmony such that
    different experiments are integrated.

    Example
    -------
    First, load libraries and example dataset, and preprocess.

    >>> import scanpy as sc
    >>> import scanpy.external as sce
    >>> adata = sc.datasets.pbmc3k()
    >>> sc.pp.recipe_zheng17(adata)
    >>> sc.pp.pca(adata)

    We now arbitrarily assign a batch metadata variable to each cell
    for the sake of example, but during real usage there would already
    be a column in ``adata.obs`` giving the experiment each cell came
    from.

    >>> adata.obs["batch"] = 1350 * ["a"] + 1350 * ["b"]

    Finally, run harmony. Afterwards, there will be a new table in
    ``adata.obsm`` containing the adjusted PC's.

    >>> sce.pp.harmony_integrate(adata, "batch")
    >>> "X_pca_harmony" in adata.obsm
    True

    """
    try:
        import harmonypy
    except ImportError as e:
        msg = "\nplease install harmonypy:\n\n\tpip install harmonypy"
        raise ImportError(msg) from e

    x = adata.obsm[basis].astype(np.float64)

    harmony_out = harmonypy.run_harmony(x, adata.obs, key, **kwargs)

    adata.obsm[adjusted_basis] = harmony_out.Z_corr.T



"""Benchmark preprocessing operations in Scanpy that run on counts.

API documentation: <https://scanpy.readthedocs.io/en/stable/api/preprocessing.html>.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import scanpy as sc

from ._utils import get_count_dataset

if TYPE_CHECKING:
    from anndata import AnnData

    from ._utils import Dataset, KeyCount

# setup variables

adata: AnnData
batch_key: str | None


def setup(dataset: Dataset, layer: KeyCount, *_):
    """Set up global variables before each benchmark."""
    global adata, batch_key
    adata, batch_key = get_count_dataset(dataset, layer=layer)
    assert "log1p" not in adata.uns


# ASV suite

params: tuple[list[Dataset], list[KeyCount]] = (
    ["pbmc68k_reduced", "pbmc3k"],
    ["counts", "counts-off-axis"],
)
param_names = ["dataset", "layer"]


def time_filter_cells(*_):
    sc.pp.filter_cells(adata, min_genes=100)


def peakmem_filter_cells(*_):
    sc.pp.filter_cells(adata, min_genes=100)


def time_filter_genes(*_):
    sc.pp.filter_genes(adata, min_cells=3)


def peakmem_filter_genes(*_):
    sc.pp.filter_genes(adata, min_cells=3)


def time_scrublet(*_):
    sc.pp.scrublet(adata, batch_key=batch_key)


def peakmem_scrublet(*_):
    sc.pp.scrublet(adata, batch_key=batch_key)


# Can’t do seurat v3 yet: https://github.com/conda-forge/scikit-misc-feedstock/issues/17
"""
def time_hvg_seurat_v3(*_):
    # seurat v3 runs on counts
    sc.pp.highly_variable_genes(adata, flavor="seurat_v3_paper")


def peakmem_hvg_seurat_v3(*_):
    sc.pp.highly_variable_genes(adata, flavor="seurat_v3_paper")
"""


class FastSuite:
    """Suite for fast preprocessing operations."""

    params: tuple[list[Dataset], list[KeyCount]] = (
        ["pbmc3k", "pbmc68k_reduced", "bmmc", "lung93k"],
        ["counts", "counts-off-axis"],
    )
    param_names = ("dataset", "layer")

    def time_calculate_qc_metrics(self, *_):
        sc.pp.calculate_qc_metrics(
            adata, qc_vars=["mt"], percent_top=None, log1p=False, inplace=True
        )

    def peakmem_calculate_qc_metrics(self, *_):
        sc.pp.calculate_qc_metrics(
            adata, qc_vars=["mt"], percent_top=None, log1p=False, inplace=True
        )

    def time_normalize_total(self, *_):
        sc.pp.normalize_total(adata, target_sum=1e4)

    def peakmem_normalize_total(self, *_):
        sc.pp.normalize_total(adata, target_sum=1e4)

    def time_log1p(self, *_):
        # TODO: This would fail: assert "log1p" not in adata.uns, "ASV bug?"
        # https://github.com/scverse/scanpy/issues/3052
        adata.uns.pop("log1p", None)
        sc.pp.log1p(adata)

    def peakmem_log1p(self, *_):
        adata.uns.pop("log1p", None)
        sc.pp.log1p(adata)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from anndata import AnnData
from anndata.tests.helpers import assert_equal
from scipy import sparse

import scanpy as sc
from scanpy._compat import CSBase
from scanpy._utils import axis_sum
from scanpy.preprocessing._normalization import _compute_nnz_median
from testing.scanpy._helpers import (
    _check_check_values_warnings,
    check_rep_mutation,
    check_rep_results,
)

# TODO: Add support for sparse-in-dask
from testing.scanpy._pytest.params import ARRAY_TYPES, ARRAY_TYPES_DENSE

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

X_total = np.array([[1, 0], [3, 0], [5, 6]])
X_frac = np.array([[1, 0, 1], [3, 0, 1], [5, 6, 1]])


@pytest.mark.parametrize("array_type", ARRAY_TYPES)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
@pytest.mark.parametrize("target_sum", [None, 1.0])
@pytest.mark.parametrize("exclude_highly_expressed", [True, False])
def test_normalize_matrix_types(
    array_type, dtype, target_sum, exclude_highly_expressed
):
    adata = sc.datasets.pbmc68k_reduced()
    adata.X = (adata.raw.X).astype(dtype)
    adata_casted = adata.copy()
    adata_casted.X = array_type(adata_casted.raw.X).astype(dtype)
    sc.pp.normalize_total(
        adata, target_sum=target_sum, exclude_highly_expressed=exclude_highly_expressed
    )
    sc.pp.normalize_total(
        adata_casted,
        target_sum=target_sum,
        exclude_highly_expressed=exclude_highly_expressed,
    )
    x = adata_casted.X
    if "dask" in array_type.__name__:
        x = x.compute()
    if isinstance(x, CSBase):
        x = x.todense()
    if isinstance(adata.X, CSBase):
        adata.X = adata.X.todense()
    np.testing.assert_allclose(x, adata.X, rtol=1e-5, atol=1e-5)


@pytest.mark.parametrize("array_type", ARRAY_TYPES)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
def test_normalize_total(array_type, dtype):
    adata = AnnData(array_type(X_total).astype(dtype))
    sc.pp.normalize_total(adata, key_added="n_counts")
    assert np.allclose(np.ravel(axis_sum(adata.X, axis=1)), [3.0, 3.0, 3.0])
    sc.pp.normalize_total(adata, target_sum=1, key_added="n_counts2")
    assert np.allclose(np.ravel(axis_sum(adata.X, axis=1)), [1.0, 1.0, 1.0])

    adata = AnnData(array_type(X_frac).astype(dtype))
    sc.pp.normalize_total(adata, exclude_highly_expressed=True, max_fraction=0.7)
    assert np.allclose(np.ravel(axis_sum(adata.X[:, 1:3], axis=1)), [1.0, 1.0, 1.0])


@pytest.mark.parametrize("array_type", ARRAY_TYPES)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
def test_normalize_total_rep(array_type, dtype):
    # Test that layer kwarg works
    x = array_type(sparse.random(100, 50, format="csr", density=0.2, dtype=dtype))
    check_rep_mutation(sc.pp.normalize_total, x, fields=["layer"])
    check_rep_results(sc.pp.normalize_total, x, fields=["layer"])


@pytest.mark.parametrize("array_type", ARRAY_TYPES)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
def test_normalize_total_view(array_type, dtype):
    adata = AnnData(array_type(X_total).astype(dtype))
    v = adata[:, :]

    sc.pp.normalize_total(v)
    sc.pp.normalize_total(adata)

    assert not v.is_view
    assert_equal(adata, v)


def test_normalize_pearson_residuals_warnings(pbmc3k_parametrized):
    adata = pbmc3k_parametrized()

    if np.issubdtype(adata.X.dtype, np.integer):
        pytest.skip("Can’t store non-integral data with int dtype")

    # depending on check_values, warnings should be raised for non-integer data
    adata_noninteger = adata.copy()
    x, y = np.nonzero(adata_noninteger.X)
    adata_noninteger.X[x[0], y[0]] = 0.5

    _check_check_values_warnings(
        function=sc.experimental.pp.normalize_pearson_residuals,
        adata=adata_noninteger,
        expected_warning="`normalize_pearson_residuals()` expects raw count data, but non-integers were found.",
    )


@pytest.mark.parametrize(
    ("params", "match"),
    [
        pytest.param(dict(theta=0), r"Pearson residuals require theta > 0", id="theta"),
        pytest.param(
            dict(theta=-1), r"Pearson residuals require theta > 0", id="theta"
        ),
        pytest.param(
            dict(clip=-1), r"Pearson residuals require `clip>=0` or `clip=None`."
        ),
    ],
)
def test_normalize_pearson_residuals_errors(pbmc3k_parametrized, params, match):
    adata = pbmc3k_parametrized()

    with pytest.raises(ValueError, match=match):
        sc.experimental.pp.normalize_pearson_residuals(adata, **params)


@pytest.mark.parametrize(
    "sparsity_func",
    [np.array, sparse.csr_matrix],  # noqa: TID251
    ids=lambda x: x.__name__,
)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
@pytest.mark.parametrize("theta", [0.01, 1, 100, np.inf])
@pytest.mark.parametrize("clip", [None, 1, np.inf])
def test_normalize_pearson_residuals_values(sparsity_func, dtype, theta, clip):
    # toy data
    x = np.array([[3, 6], [2, 4], [1, 0]])
    ns = np.sum(x, axis=1)
    ps = np.sum(x, axis=0) / np.sum(x)
    mu = np.outer(ns, ps)

    # compute reference residuals
    if np.isinf(theta):
        # Poisson case
        residuals_reference = (x - mu) / np.sqrt(mu)
    else:
        # NB case
        residuals_reference = (x - mu) / np.sqrt(mu + mu**2 / theta)

    # compute output to test
    adata = AnnData(sparsity_func(x).astype(dtype))
    output = sc.experimental.pp.normalize_pearson_residuals(
        adata, theta=theta, clip=clip, inplace=False
    )
    output_x = output["X"]
    sc.experimental.pp.normalize_pearson_residuals(
        adata, theta=theta, clip=clip, inplace=True
    )

    # check for correct new `adata.uns` keys
    assert {"pearson_residuals_normalization"} <= adata.uns.keys()
    assert {"theta", "clip", "computed_on"} <= adata.uns[
        "pearson_residuals_normalization"
    ].keys()
    # test against inplace
    np.testing.assert_array_equal(adata.X, output_x)

    if clip is None:
        # default clipping: compare to sqrt(n) threshold
        clipping_threshold = np.sqrt(adata.shape[0]).astype(np.float32)
        assert np.max(output_x) <= clipping_threshold
        assert np.min(output_x) >= -clipping_threshold
    elif np.isinf(clip):
        # no clipping: compare to raw residuals
        assert np.allclose(output_x, residuals_reference)
    else:
        # custom clipping: compare to custom threshold
        assert np.max(output_x) <= clip
        assert np.min(output_x) >= -clip


def _check_pearson_pca_fields(ad, n_cells, n_comps):
    assert {"pearson_residuals_normalization", "pca"} <= ad.uns.keys(), (
        "Missing `.uns` keys. Expected `['pearson_residuals_normalization', 'pca']`, "
        f"but only {list(ad.uns.keys())} were found"
    )
    assert "X_pca" in ad.obsm, (
        f"Missing `obsm` key `'X_pca'`, only {list(ad.obsm.keys())} were found"
    )
    assert "PCs" in ad.varm, (
        f"Missing `varm` key `'PCs'`, only {list(ad.varm.keys())} were found"
    )
    assert ad.obsm["X_pca"].shape == (
        n_cells,
        n_comps,
    ), "Wrong shape of PCA output in `X_pca`"


@pytest.mark.parametrize("n_hvgs", [100, 200])
@pytest.mark.parametrize("n_comps", [30, 50])
@pytest.mark.parametrize(
    ("do_hvg", "params", "n_var_copy_name"),
    [
        pytest.param(False, dict(), "n_genes", id="no_hvg"),
        pytest.param(True, dict(), "n_hvgs", id="hvg_default"),
        pytest.param(
            True, dict(use_highly_variable=False), "n_genes", id="hvg_opt_out"
        ),
        pytest.param(False, dict(mask_var="test_mask"), "n_unmasked", id="mask"),
    ],
)
def test_normalize_pearson_residuals_pca(
    *,
    pbmc3k_parametrized_small: Callable[[], AnnData],
    n_hvgs: int,
    n_comps: int,
    do_hvg: bool,
    params: dict[str, Any],
    n_var_copy_name: str,  # number of variables in output if inplace=False
):
    adata = pbmc3k_parametrized_small()
    n_cells, n_genes = adata.shape
    n_unmasked = n_genes - 5
    adata.var["test_mask"] = np.r_[
        np.repeat(True, n_unmasked), np.repeat(False, n_genes - n_unmasked)  # noqa: FBT003
    ]
    n_var_copy = locals()[n_var_copy_name]
    assert isinstance(n_var_copy, int | np.integer)

    if do_hvg:
        sc.experimental.pp.highly_variable_genes(
            adata, flavor="pearson_residuals", n_top_genes=n_hvgs
        )

    # inplace=False
    adata_pca = sc.experimental.pp.normalize_pearson_residuals_pca(
        adata.copy(), inplace=False, n_comps=n_comps, **params
    )
    # inplace=True modifies the input adata object
    sc.experimental.pp.normalize_pearson_residuals_pca(
        adata, inplace=True, n_comps=n_comps, **params
    )

    for ad, n_var_ret in (
        (adata_pca, n_var_copy),
        # inplace adatas should always retains original shape
        (adata, n_genes),
    ):
        _check_pearson_pca_fields(ad, n_cells, n_comps)

        # check adata shape to see if all genes or only HVGs are in the returned adata
        assert ad.shape == (n_cells, n_var_ret)

        # check PC shapes to see whether or not HVGs were used for PCA
        assert ad.varm["PCs"].shape == (n_var_ret, n_comps)

    # check if there are columns of all-zeros in the PCs shapes
    # to see whether or not HVGs were used for PCA
    # either no all-zero-colums or all number corresponding to non-hvgs should exist
    assert sum(np.sum(np.abs(adata.varm["PCs"]), axis=1) == 0) == (n_genes - n_var_copy)

    # compare PCA results beteen inplace / copied
    np.testing.assert_array_equal(adata.obsm["X_pca"], adata_pca.obsm["X_pca"])


@pytest.mark.parametrize("n_hvgs", [100, 200])
@pytest.mark.parametrize("n_comps", [30, 50])
def test_normalize_pearson_residuals_recipe(pbmc3k_parametrized_small, n_hvgs, n_comps):
    adata = pbmc3k_parametrized_small()
    n_cells, n_genes = adata.shape

    ### inplace = False ###
    # outputs the (potentially hvg-restricted) adata_pca object
    # PCA on all genes
    adata_pca, hvg = sc.experimental.pp.recipe_pearson_residuals(
        adata.copy(), inplace=False, n_comps=n_comps, n_top_genes=n_hvgs
    )

    # check PCA fields
    _check_pearson_pca_fields(adata_pca, n_cells, n_comps)
    # check adata output shape (only HVGs in output)
    assert adata_pca.shape == (n_cells, n_hvgs)
    # check PC shape (non-hvgs are removed, so only `n_hvgs` genes)
    assert adata_pca.varm["PCs"].shape == (n_hvgs, n_comps)

    # check hvg df
    assert {
        "means",
        "variances",
        "residual_variances",
        "highly_variable_rank",
        "highly_variable",
    } <= set(hvg.columns)
    assert np.sum(hvg["highly_variable"]) == n_hvgs
    assert hvg.shape[0] == n_genes

    ### inplace = True ###
    # modifies the input adata object
    # PCA on all genes
    sc.experimental.pp.recipe_pearson_residuals(
        adata, inplace=True, n_comps=n_comps, n_top_genes=n_hvgs
    )

    # check PCA fields and output shape
    _check_pearson_pca_fields(adata, n_cells, n_comps)
    # check adata shape (no change to input)
    assert adata.shape == (n_cells, n_genes)
    # check PC shape (non-hvgs are masked with 0s, so original number of genes)
    assert adata.varm["PCs"].shape == (n_genes, n_comps)
    # number of all-zero-colums should be number of non-hvgs
    assert sum(np.sum(np.abs(adata.varm["PCs"]), axis=1) == 0) == n_genes - n_hvgs


@pytest.mark.parametrize("array_type", ARRAY_TYPES_DENSE)
@pytest.mark.parametrize("dtype", ["float32", "int64"])
def test_compute_nnz_median(array_type, dtype):
    data = np.array([0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=dtype)
    data = array_type(data)
    np.testing.assert_allclose(_compute_nnz_median(data), 5)



"""Some common fixtures for use in tests.

This is kept seperate from the helpers file because it relies on pytest.
"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import pytest

from .data import (
    backed_adata,
    pbmc3k_parametrized,
    pbmc3k_parametrized_small,
    pbmc3ks_parametrized_session,
)

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

__all__ = [
    "_doctest_env",
    "backed_adata",
    "float_dtype",
    "pbmc3k_parametrized",
    "pbmc3k_parametrized_small",
    "pbmc3ks_parametrized_session",
]


@pytest.fixture(params=[np.float64, np.float32])
def float_dtype(request):
    return request.param


@pytest.fixture
def _doctest_env(cache: pytest.Cache, tmp_path: Path) -> Generator[None, None, None]:
    from scanpy._compat import chdir

    showwarning_orig = warnings.showwarning

    def showwarning(message, category, filename, lineno, file=None, line=None) -> None:  # noqa: PLR0917
        if file is None:
            if line is None:
                import linecache

                line = linecache.getline(filename, lineno)
            line = line.strip()
            print(f"{category.__name__}: {message}\n    {line}")
        else:
            showwarning_orig(message, category, filename, lineno, file, line)

    # make errors visible and the rest ignored
    warnings.filters = [
        ("default", *rest) for action, *rest in warnings.filters if action == "error"
    ] + [("ignore", None, Warning, None, 0)]

    warnings.showwarning = showwarning
    with chdir(tmp_path):
        yield
    warnings.showwarning = showwarning_orig



"""Extension to patch https://github.com/executablebooks/MyST-NB/pull/599."""

# TODO once MyST-NB 1.1.1/1.2.0 is out, this can be removed.

from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING

from myst_nb.core.render import MditRenderMixin

if TYPE_CHECKING:
    from sphinx.application import Sphinx


get_orig = MditRenderMixin.get_cell_level_config


def get_cell_level_config(
    self: MditRenderMixin,
    field: str,
    cell_metadata: dict[str, object],
    line: int | None = None,
):
    """Correct version of ``MditRenderMixin.get_cell_level_config``."""
    rv = get_orig(self, field, cell_metadata, line)
    return copy(rv)


def setup(app: Sphinx):
    """App setup hook."""
    MditRenderMixin.get_cell_level_config = get_cell_level_config



"""Embed high-dimensional data using PHATE."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ... import logging as logg
from ..._compat import old_positionals
from ..._settings import settings
from ..._utils._doctests import doctest_needs

if TYPE_CHECKING:
    from typing import Literal

    from anndata import AnnData

    from ..._utils.random import _LegacyRandom


@old_positionals(
    "k",
    "a",
    "n_landmark",
    "t",
    "gamma",
    "n_pca",
    "knn_dist",
    "mds_dist",
    "mds",
    "n_jobs",
    "random_state",
    "verbose",
    "copy",
)
@doctest_needs("phate")
def phate(  # noqa: PLR0913
    adata: AnnData,
    n_components: int = 2,
    *,
    k: int = 5,
    a: int = 15,
    n_landmark: int = 2000,
    t: int | str = "auto",
    gamma: float = 1.0,
    n_pca: int = 100,
    knn_dist: str = "euclidean",
    mds_dist: str = "euclidean",
    mds: Literal["classic", "metric", "nonmetric"] = "metric",
    n_jobs: int | None = None,
    random_state: _LegacyRandom = None,
    verbose: bool | int | None = None,
    copy: bool = False,
    **kwargs,
) -> AnnData | None:
    """PHATE :cite:p:`Moon2019`.

    Potential of Heat-diffusion for Affinity-based Trajectory Embedding (PHATE)
    embeds high dimensional single-cell data into two or three dimensions for
    visualization of biological progressions.

    For more information and access to the object-oriented interface, read the
    `PHATE documentation <https://phate.readthedocs.io/>`__.  For
    tutorials, bug reports, and R/MATLAB implementations, visit the `PHATE
    GitHub page <https://github.com/KrishnaswamyLab/PHATE/>`__. For help
    using PHATE, go `here <https://krishnaswamylab.org/get-help>`__.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_components
        number of dimensions in which the data will be embedded
    k
        number of nearest neighbors on which to build kernel
    a
        sets decay rate of kernel tails.
        If None, alpha decaying kernel is not used
    n_landmark
        number of landmarks to use in fast PHATE
    t
        power to which the diffusion operator is powered
        sets the level of diffusion. If 'auto', t is selected
        according to the knee point in the Von Neumann Entropy of
        the diffusion operator
    gamma
        Informational distance constant between -1 and 1.
        `gamma=1` gives the PHATE log potential, `gamma=0` gives
        a square root potential.
    n_pca
        Number of principal components to use for calculating
        neighborhoods. For extremely large datasets, using
        n_pca < 20 allows neighborhoods to be calculated in
        log(n_samples) time.
    knn_dist
        recommended values: 'euclidean' and 'cosine'
        Any metric from `scipy.spatial.distance` can be used
        distance metric for building kNN graph
    mds_dist
        recommended values: 'euclidean' and 'cosine'
        Any metric from `scipy.spatial.distance` can be used
        distance metric for MDS
    mds
        Selects which MDS algorithm is used for dimensionality reduction.
    n_jobs
        The number of jobs to use for the computation.
        If `None`, `sc.settings.n_jobs` is used.
        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging.
        For n_jobs below -1, (n_cpus + 1 + n_jobs) are used. Thus for
        n_jobs = -2, all CPUs but one are used
    random_state
        Random seed. Defaults to the global `numpy` random number generator
    verbose
        If `True` or an :class:`int`/:class:`~scanpy.Verbosity` ≥ 2/:attr:`~scanpy.Verbosity.hint`, print status messages.
        If `None`, :attr:`scanpy.settings.verbosity` is used.
    copy
        Return a copy instead of writing to `adata`.
    kwargs
        Additional arguments to `phate.PHATE`

    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.

    **X_phate** : `np.ndarray`, (`adata.obs`, shape=[n_samples, n_components], dtype `float`)
        PHATE coordinates of data.

    Examples
    --------
    >>> from anndata import AnnData
    >>> import scanpy.external as sce
    >>> import phate
    >>> tree_data, tree_clusters = phate.tree.gen_dla(
    ...     n_dim=100,
    ...     n_branch=20,
    ...     branch_length=100,
    ... )
    >>> tree_data.shape
    (2000, 100)
    >>> adata = AnnData(tree_data)
    >>> sce.tl.phate(adata, k=5, a=20, t=150)
    >>> adata.obsm["X_phate"].shape
    (2000, 2)
    >>> sce.pl.phate(adata)

    """
    start = logg.info("computing PHATE")
    adata = adata.copy() if copy else adata
    verbosity = settings.verbosity if verbose is None else verbose
    verbose = verbosity if isinstance(verbosity, bool) else verbosity >= 2
    n_jobs = settings.n_jobs if n_jobs is None else n_jobs
    try:
        import phate
    except ImportError as e:
        msg = (
            "You need to install the package `phate`: please run `pip install "
            "--user phate` in a terminal."
        )
        raise ImportError(msg) from e
    x_phate = phate.PHATE(
        n_components=n_components,
        k=k,
        a=a,
        n_landmark=n_landmark,
        t=t,
        gamma=gamma,
        n_pca=n_pca,
        knn_dist=knn_dist,
        mds_dist=mds_dist,
        mds=mds,
        n_jobs=n_jobs,
        random_state=random_state,
        verbose=verbose,
        **kwargs,
    ).fit_transform(adata)
    # update AnnData instance
    adata.obsm["X_phate"] = x_phate  # annotate samples with PHATE coordinates
    logg.info(
        "    finished",
        time=start,
        deep=("added\n    'X_phate', PHATE coordinates (adata.obsm)"),
    )
    return adata if copy else None



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import TYPE_CHECKING, cast

import numpy as np
import pandas as pd
from anndata import AnnData, concat
from scipy import sparse
from sklearn.utils import check_random_state

from ... import logging as logg
from ...neighbors import (
    Neighbors,
    _get_indices_distances_from_sparse_matrix,
)
from .._utils import sample_comb
from .sparse_utils import subsample_counts

if TYPE_CHECKING:
    from numpy.random import RandomState
    from numpy.typing import NDArray

    from ..._compat import CSBase, CSCBase
    from ..._utils.random import _LegacyRandom
    from ...neighbors import _Metric, _MetricFn

__all__ = ["Scrublet"]


@dataclass(kw_only=True)
class Scrublet:
    """Initialize Scrublet object with counts matrix and doublet prediction parameters.

    Parameters
    ----------
    counts_obs
        Matrix with shape (n_cells, n_genes) containing raw (unnormalized)
        UMI-based transcript counts.
        Converted into a :class:`scipy.sparse.csc_matrix`.

    total_counts_obs
        Array with shape (n_cells,) of total UMI counts per cell.
        If `None`, this is calculated as the row sums of `counts_obs`.

    sim_doublet_ratio
        Number of doublets to simulate relative to the number of observed
        transcriptomes.

    n_neighbors
        Number of neighbors used to construct the KNN graph of observed
        transcriptomes and simulated doublets.
        If `None`, this is set to round(0.5 * sqrt(n_cells))

    expected_doublet_rate
        The estimated doublet rate for the experiment.

    stdev_doublet_rate
        Uncertainty in the expected doublet rate.

    random_state
        Random state for doublet simulation, approximate
        nearest neighbor search, and PCA/TruncatedSVD.

    """

    # init fields

    counts_obs: InitVar[CSBase | NDArray[np.integer]] = field(kw_only=False)
    total_counts_obs: InitVar[NDArray[np.integer] | None] = None
    sim_doublet_ratio: float = 2.0
    n_neighbors: InitVar[int | None] = None
    expected_doublet_rate: float = 0.1
    stdev_doublet_rate: float = 0.02
    random_state: InitVar[_LegacyRandom] = 0

    # private fields

    _n_neighbors: int = field(init=False, repr=False)
    _random_state: RandomState = field(init=False, repr=False)

    _counts_obs: CSCBase = field(init=False, repr=False)
    _total_counts_obs: NDArray[np.integer] = field(init=False, repr=False)
    _counts_obs_norm: CSBase = field(init=False, repr=False)

    _counts_sim: CSBase = field(init=False, repr=False)
    _total_counts_sim: NDArray[np.integer] = field(init=False, repr=False)
    _counts_sim_norm: CSBase | None = field(default=None, init=False, repr=False)

    # Fields set by methods

    predicted_doublets_: NDArray[np.bool_] | None = field(init=False)
    """(shape: n_cells)
    Boolean mask of predicted doublets in the observed transcriptomes.
    """

    doublet_scores_obs_: NDArray[np.float64] = field(init=False)
    """(shape: n_cells)
    Doublet scores for observed transcriptomes.
    """

    doublet_scores_sim_: NDArray[np.float64] = field(init=False)
    """(shape: n_doublets)
    Doublet scores for simulated doublets.
    """

    doublet_errors_obs_: NDArray[np.float64] = field(init=False)
    """(shape: n_cells)
    Standard error in the doublet scores for observed transcriptomes.
    """

    doublet_errors_sim_: NDArray[np.float64] = field(init=False)
    """(shape: n_doublets)
    Standard error in the doublet scores for simulated doublets.
    """

    threshold_: float = field(init=False)
    """Doublet score threshold for calling a transcriptome a doublet."""

    z_scores_: NDArray[np.float64] = field(init=False)
    """(shape: n_cells)
    Z-score conveying confidence in doublet calls.
    Z = `(doublet_score_obs_ - threhsold_) / doublet_errors_obs_`
    """

    detected_doublet_rate_: float = field(init=False)
    """Fraction of observed transcriptomes that have been called doublets."""

    detectable_doublet_fraction_: float = field(init=False)
    """Estimated fraction of doublets that are detectable, i.e.,
    fraction of simulated doublets with doublet scores above `threshold_`
    """

    overall_doublet_rate_: float = field(init=False)
    """Estimated overall doublet rate,
    `detected_doublet_rate_ / detectable_doublet_fraction_`.
    Should agree (roughly) with `expected_doublet_rate`.
    """

    manifold_obs_: NDArray[np.float64] = field(init=False)
    """(shape: n_cells × n_features)
    The single-cell "manifold" coordinates (e.g., PCA coordinates)
    for observed transcriptomes. Nearest neighbors are found using
    the union of `manifold_obs_` and `manifold_sim_` (see below).
    """

    manifold_sim_: NDArray[np.float64] = field(init=False)
    """shape (n_doublets × n_features)
    The single-cell "manifold" coordinates (e.g., PCA coordinates)
    for simulated doublets. Nearest neighbors are found using
    the union of `manifold_obs_` (see above) and `manifold_sim_`.
    """

    doublet_parents_: NDArray[np.intp] = field(init=False)
    """(shape: n_doublets × 2)
    Indices of the observed transcriptomes used to generate the
    simulated doublets.
    """

    doublet_neighbor_parents_: list[NDArray[np.intp]] = field(init=False)
    """(length: n_cells)
    A list of arrays of the indices of the doublet neighbors of
    each observed transcriptome (the ith entry is an array of
    the doublet neighbors of transcriptome i).
    """

    def __post_init__(
        self,
        counts_obs: CSBase | NDArray[np.integer],
        total_counts_obs: NDArray[np.integer] | None,
        n_neighbors: int | None,
        random_state: _LegacyRandom,
    ) -> None:
        self._counts_obs = sparse.csc_matrix(counts_obs)  # noqa: TID251
        self._total_counts_obs = (
            np.asarray(self._counts_obs.sum(1)).squeeze()
            if total_counts_obs is None
            else total_counts_obs
        )
        self._n_neighbors = (
            round(0.5 * np.sqrt(self._counts_obs.shape[0]))
            if n_neighbors is None
            else n_neighbors
        )
        self._random_state = check_random_state(random_state)

    def simulate_doublets(
        self,
        *,
        sim_doublet_ratio: float | None = None,
        synthetic_doublet_umi_subsampling: float = 1.0,
    ) -> None:
        """Simulate doublets by adding the counts of random observed transcriptome pairs.

        Parameters
        ----------
        sim_doublet_ratio
            Number of doublets to simulate relative to the number of observed
            transcriptomes. If `None`, self.sim_doublet_ratio is used.

        synthetic_doublet_umi_subsampling
            Rate for sampling UMIs when creating synthetic doublets.
            If 1.0, each doublet is created by simply adding the UMIs from two randomly
            sampled observed transcriptomes.
            For values less than 1, the UMI counts are added and then randomly sampled
            at the specified rate.

        Sets
        ----
        doublet_parents_

        """
        if sim_doublet_ratio is None:
            sim_doublet_ratio = self.sim_doublet_ratio
        else:
            self.sim_doublet_ratio = sim_doublet_ratio

        n_obs = self._counts_obs.shape[0]
        n_sim = int(n_obs * sim_doublet_ratio)

        pair_ix = sample_comb((n_obs, n_obs), n_sim, random_state=self._random_state)

        e1 = cast("CSCBase", self._counts_obs[pair_ix[:, 0], :])
        e2 = cast("CSCBase", self._counts_obs[pair_ix[:, 1], :])
        tots1 = self._total_counts_obs[pair_ix[:, 0]]
        tots2 = self._total_counts_obs[pair_ix[:, 1]]
        if synthetic_doublet_umi_subsampling < 1:
            self._counts_sim, self._total_counts_sim = subsample_counts(
                e1 + e2,
                rate=synthetic_doublet_umi_subsampling,
                original_totals=tots1 + tots2,
                random_seed=self._random_state,
            )
        else:
            self._counts_sim = e1 + e2
            self._total_counts_sim = tots1 + tots2
        self.doublet_parents_ = pair_ix

    def set_manifold(
        self, manifold_obs: NDArray[np.float64], manifold_sim: NDArray[np.float64]
    ) -> None:
        """Set the manifold coordinates used in k-nearest-neighbor graph construction.

        Parameters
        ----------
        manifold_obs
            (shape: n_cells × n_features)
            The single-cell "manifold" coordinates (e.g., PCA coordinates)
            for observed transcriptomes. Nearest neighbors are found using
            the union of `manifold_obs` and `manifold_sim` (see below).

        manifold_sim
            (shape: n_doublets × n_features)
            The single-cell "manifold" coordinates (e.g., PCA coordinates)
            for simulated doublets. Nearest neighbors are found using
            the union of `manifold_obs` (see above) and `manifold_sim`.

        Sets
        ----
        manifold_obs_, manifold_sim_,

        """
        self.manifold_obs_ = manifold_obs
        self.manifold_sim_ = manifold_sim

    def calculate_doublet_scores(
        self,
        *,
        use_approx_neighbors: bool | None = None,
        distance_metric: _Metric | _MetricFn = "euclidean",
        get_doublet_neighbor_parents: bool = False,
    ) -> NDArray[np.float64]:
        """Calculate doublet scores for observed transcriptomes and simulated doublets.

        Requires that manifold_obs_ and manifold_sim_ have already been set.

        Parameters
        ----------
        use_approx_neighbors
            Use approximate nearest neighbor method (annoy) for the KNN
            classifier.

        distance_metric
            Distance metric used when finding nearest neighbors. For list of
            valid values, see the documentation for annoy (if `use_approx_neighbors`
            is True) or sklearn.neighbors.NearestNeighbors (if `use_approx_neighbors`
            is False).

        get_doublet_neighbor_parents
            If True, return the parent transcriptomes that generated the
            doublet neighbors of each observed transcriptome. This information can
            be used to infer the cell states that generated a given
            doublet state.

        Sets
        ----
        doublet_scores_obs_, doublet_scores_sim_,
        doublet_errors_obs_, doublet_errors_sim_,
        doublet_neighbor_parents_

        """
        self._nearest_neighbor_classifier(
            k=self._n_neighbors,
            exp_doub_rate=self.expected_doublet_rate,
            stdev_doub_rate=self.stdev_doublet_rate,
            use_approx_neighbors=use_approx_neighbors,
            distance_metric=distance_metric,
            get_neighbor_parents=get_doublet_neighbor_parents,
        )
        return self.doublet_scores_obs_

    def _nearest_neighbor_classifier(
        self,
        k: int = 40,
        *,
        use_approx_neighbors: bool | None = None,
        distance_metric: _Metric | _MetricFn = "euclidean",
        exp_doub_rate: float = 0.1,
        stdev_doub_rate: float = 0.03,
        get_neighbor_parents: bool = False,
    ) -> None:
        adatas = [
            AnnData(
                (arr := getattr(self, f"manifold_{n}_")),
                obs=dict(
                    obs_names=pd.RangeIndex(arr.shape[0]).astype("string") + n,
                    doub_labels=n,
                ),
            )
            for n in ["obs", "sim"]
        ]
        manifold = concat(adatas)

        n_obs: int = (manifold.obs["doub_labels"] == "obs").sum()
        n_sim: int = (manifold.obs["doub_labels"] == "sim").sum()

        # Adjust k (number of nearest neighbors) based on the ratio of simulated to observed cells
        k_adj = round(k * (1 + n_sim / float(n_obs)))

        # Find k_adj nearest neighbors
        knn = Neighbors(manifold)
        transformer = None
        if use_approx_neighbors is not None:
            transformer = "pynndescent" if use_approx_neighbors else "sklearn"
        knn.compute_neighbors(
            k_adj,
            metric=distance_metric,
            knn=True,
            transformer=transformer,
            method=None,
            random_state=self._random_state,
        )
        neighbors, _ = _get_indices_distances_from_sparse_matrix(knn.distances, k_adj)
        if use_approx_neighbors:
            neighbors = neighbors[:, 1:]
        # Calculate doublet score based on ratio of simulated cell neighbors vs. observed cell neighbors
        doub_neigh_mask: NDArray[np.bool_] = (
            manifold.obs["doub_labels"].to_numpy()[neighbors] == "sim"
        )
        n_sim_neigh: NDArray[np.int64] = doub_neigh_mask.sum(axis=1)

        rho = exp_doub_rate
        r = n_sim / float(n_obs)
        nd = n_sim_neigh.astype(np.float64)
        n = float(k_adj)

        # Bayesian
        q = (nd + 1) / (n + 2)
        ld = q * rho / r / (1 - rho - q * (1 - rho - rho / r))

        se_q = np.sqrt(q * (1 - q) / (n + 3))
        se_rho = stdev_doub_rate

        se_ld = (
            q
            * rho
            / r
            / (1 - rho - q * (1 - rho - rho / r)) ** 2
            * np.sqrt((se_q / q * (1 - rho)) ** 2 + (se_rho / rho * (1 - q)) ** 2)
        )

        self.doublet_scores_obs_ = ld[manifold.obs["doub_labels"] == "obs"]
        self.doublet_scores_sim_ = ld[manifold.obs["doub_labels"] == "sim"]
        self.doublet_errors_obs_ = se_ld[manifold.obs["doub_labels"] == "obs"]
        self.doublet_errors_sim_ = se_ld[manifold.obs["doub_labels"] == "sim"]

        # get parents of doublet neighbors, if requested
        neighbor_parents = None
        if get_neighbor_parents:
            parent_cells = self.doublet_parents_
            neighbors = neighbors - n_obs
            neighbor_parents = []
            for c in range(n_obs):
                this_doub_neigh = neighbors[c, :][neighbors[c, :] > -1]
                if len(this_doub_neigh) > 0:
                    this_doub_neigh_parents = np.unique(
                        parent_cells[this_doub_neigh, :].flatten()
                    )
                    neighbor_parents.append(this_doub_neigh_parents)
                else:
                    neighbor_parents.append(np.array([], dtype=np.intp))
            self.doublet_neighbor_parents_ = neighbor_parents

    def call_doublets(
        self, *, threshold: float | None = None, verbose: bool = True
    ) -> NDArray[np.bool_] | None:
        """Call trancriptomes as doublets or singlets.

        Parameters
        ----------
        threshold
            Doublet score threshold for calling a transcriptome
            a doublet. If `None`, this is set automatically by looking
            for the minimum between the two modes of the `doublet_scores_sim_`
            histogram. It is best practice to check the threshold visually
            using the `doublet_scores_sim_` histogram and/or based on
            co-localization of predicted doublets in a 2-D embedding.

        verbose
            If True, log summary statistics.

        Sets
        ----
        predicted_doublets_, z_scores_, threshold_,
        detected_doublet_rate_, detectable_doublet_fraction,
        overall_doublet_rate_

        """
        if threshold is None:
            # automatic threshold detection
            # http://scikit-image.org/docs/dev/api/skimage.filters.html
            from skimage.filters import threshold_minimum

            try:
                threshold = cast("float", threshold_minimum(self.doublet_scores_sim_))
                if verbose:
                    logg.info(
                        f"Automatically set threshold at doublet score = {threshold:.2f}"
                    )
            except Exception:  # noqa: BLE001
                self.predicted_doublets_ = None
                if verbose:
                    logg.warning(
                        "Failed to automatically identify doublet score threshold. "
                        "Run `call_doublets` with user-specified threshold."
                    )
                return self.predicted_doublets_

        ld_obs = self.doublet_scores_obs_
        ld_sim = self.doublet_scores_sim_
        se_obs = self.doublet_errors_obs_
        z = (ld_obs - threshold) / se_obs
        self.predicted_doublets_ = ld_obs > threshold
        self.z_scores_ = z
        self.threshold_ = threshold
        self.detected_doublet_rate_ = (ld_obs > threshold).sum() / float(len(ld_obs))
        self.detectable_doublet_fraction_ = (ld_sim > threshold).sum() / float(
            len(ld_sim)
        )
        self.overall_doublet_rate_ = (
            self.detected_doublet_rate_ / self.detectable_doublet_fraction_
        )

        if verbose:
            logg.info(
                f"Detected doublet rate = {100 * self.detected_doublet_rate_:.1f}%\n"
                f"Estimated detectable doublet fraction = {100 * self.detectable_doublet_fraction_:.1f}%\n"
                "Overall doublet rate:\n"
                f"\tExpected   = {100 * self.expected_doublet_rate:.1f}%\n"
                f"\tEstimated  = {100 * self.overall_doublet_rate_:.1f}%"
            )

        return self.predicted_doublets_



from __future__ import annotations

import numpy as np
import pandas as pd
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib import rcParams

from .._compat import deprecated, old_positionals
from .._settings import settings
from ._utils import savefig_or_show

# --------------------------------------------------------------------------------
# Plot result of preprocessing functions
# --------------------------------------------------------------------------------


@old_positionals("log", "show", "save", "highly_variable_genes")
def highly_variable_genes(  # noqa: PLR0912
    adata_or_result: AnnData | pd.DataFrame | np.recarray,
    *,
    log: bool = False,
    show: bool | None = None,
    save: bool | str | None = None,
    highly_variable_genes: bool = True,
) -> None:
    """Plot dispersions or normalized variance versus means for genes.

    Produces Supp. Fig. 5c of Zheng et al. (2017) and MeanVarPlot() and
    VariableFeaturePlot() of Seurat.

    Parameters
    ----------
    adata
        Result of :func:`~scanpy.pp.highly_variable_genes`.
    log
        Plot on logarithmic axes.
    show
         Show the plot, do not return axis.
    save
        If `True` or a `str`, save the figure.
        A string is appended to the default filename.
        Infer the filetype if ending on {{`'.pdf'`, `'.png'`, `'.svg'`}}.

    """
    if isinstance(adata_or_result, AnnData):
        result = adata_or_result.var
        seurat_v3_flavor = adata_or_result.uns["hvg"]["flavor"] == "seurat_v3"
    else:
        result = adata_or_result
        if isinstance(result, pd.DataFrame):
            seurat_v3_flavor = "variances_norm" in result.columns
        else:
            seurat_v3_flavor = False
    if highly_variable_genes:
        gene_subset = result.highly_variable
    else:
        gene_subset = result.gene_subset
    means = result.means

    if seurat_v3_flavor:
        var_or_disp = result.variances
        var_or_disp_norm = result.variances_norm
    else:
        var_or_disp = result.dispersions
        var_or_disp_norm = result.dispersions_norm
    size = rcParams["figure.figsize"]
    plt.figure(figsize=(2 * size[0], size[1]))
    plt.subplots_adjust(wspace=0.3)
    for idx, d in enumerate([var_or_disp_norm, var_or_disp]):
        plt.subplot(1, 2, idx + 1)
        for label, color, mask in zip(
            ["highly variable genes", "other genes"],
            ["black", "grey"],
            [gene_subset, ~gene_subset],
            strict=True,
        ):
            if False:
                means_, var_or_disps_ = np.log10(means[mask]), np.log10(d[mask])
            else:
                means_, var_or_disps_ = means[mask], d[mask]
            plt.scatter(means_, var_or_disps_, label=label, c=color, s=1)
        if log:  # there's a bug in autoscale
            plt.xscale("log")
            plt.yscale("log")
            y_min = np.min(var_or_disp)
            y_min = 0.95 * y_min if y_min > 0 else 1e-1
            plt.xlim(0.95 * np.min(means), 1.05 * np.max(means))
            plt.ylim(y_min, 1.05 * np.max(var_or_disp))
        if idx == 0:
            plt.legend()
        plt.xlabel(f"{'$log_{10}$ ' if False else ''}mean expressions of genes")
        data_type = "dispersions" if not seurat_v3_flavor else "variances"
        plt.ylabel(
            f"{'$log_{10}$ ' if False else ''}{data_type} "
            f"of genes ({'' if idx == 0 else 'not '}normalized)"
        )

    show = settings.autoshow if show is None else show
    savefig_or_show("filter_genes_dispersion", show=show, save=save)
    if show:
        return None
    return plt.gca()


# backwards compat
@deprecated("Use sc.pl.highly_variable_genes instead")
@old_positionals("log", "show", "save")
def filter_genes_dispersion(
    result: np.recarray,
    *,
    log: bool = False,
    show: bool | None = None,
    save: bool | str | None = None,
) -> None:
    """Plot dispersions versus means for genes.

    Produces Supp. Fig. 5c of Zheng et al. (2017) and MeanVarPlot() of Seurat.

    Parameters
    ----------
    result
        Result of :func:`~scanpy.pp.filter_genes_dispersion`.
    log
        Plot on logarithmic axes.
    show
         Show the plot, do not return axis.
    save
        If `True` or a `str`, save the figure.
        A string is appended to the default filename.
        Infer the filetype if ending on {{`'.pdf'`, `'.png'`, `'.svg'`}}.

    """
    highly_variable_genes(
        result, log=log, show=show, save=save, highly_variable_genes=False
    )



from __future__ import annotations

import random
from collections.abc import Sequence
from contextlib import contextmanager
from functools import WRAPPER_ASSIGNMENTS, wraps
from typing import TYPE_CHECKING

import numpy as np
from sklearn.utils import check_random_state

from . import ensure_igraph

if TYPE_CHECKING:
    from collections.abc import Generator

    from numpy.typing import NDArray


__all__ = [
    "RNGLike",
    "SeedLike",
    "_LegacyRandom",
    "ith_k_tuple",
    "legacy_numpy_gen",
    "random_k_tuples",
    "random_str",
]

SeedLike = int | np.integer | Sequence[int] | np.random.SeedSequence
RNGLike = np.random.Generator | np.random.BitGenerator
_LegacyRandom = int | np.random.RandomState | None


###################################
# Compatibility with igraph’s RNG #
###################################


class _RNGIgraph:
    """Random number generator for igraph so global seed is not changed.

    See :func:`igraph.set_random_number_generator` for the requirements.
    """

    def __init__(self, random_state: int | np.random.RandomState = 0) -> None:
        self._rng = check_random_state(random_state)

    def getrandbits(self, k: int) -> int:
        return self._rng.tomaxint() & ((1 << k) - 1)

    def randint(self, a: int, b: int) -> int:
        return self._rng.randint(a, b + 1)

    def __getattr__(self, attr: str):
        return getattr(self._rng, "normal" if attr == "gauss" else attr)


@contextmanager
def set_igraph_random_state(
    random_state: int | np.random.RandomState,
) -> Generator[None, None, None]:
    ensure_igraph()
    import igraph

    rng = _RNGIgraph(random_state)
    try:
        igraph.set_random_number_generator(rng)
        yield None
    finally:
        igraph.set_random_number_generator(random)


###################################
# Compatibility with legacy numpy #
###################################


def legacy_numpy_gen(
    random_state: _LegacyRandom | None = None,
) -> np.random.Generator:
    """Return a random generator that behaves like the legacy one."""
    if random_state is not None:
        if isinstance(random_state, np.random.RandomState):
            np.random.set_state(random_state.get_state(legacy=False))
            return _FakeRandomGen(random_state)
        np.random.seed(random_state)
    return _FakeRandomGen(np.random.RandomState(np.random.get_bit_generator()))


class _FakeRandomGen(np.random.Generator):
    _state: np.random.RandomState

    def __init__(self, random_state: np.random.RandomState) -> None:
        self._state = random_state

    @classmethod
    def _delegate(cls) -> None:
        for name, meth in np.random.Generator.__dict__.items():
            if name.startswith("_") or not callable(meth):
                continue

            def mk_wrapper(name: str, meth):
                # Old pytest versions try to run the doctests
                @wraps(meth, assigned=set(WRAPPER_ASSIGNMENTS) - {"__doc__"})
                def wrapper(self: _FakeRandomGen, *args, **kwargs):
                    return getattr(self._state, name)(*args, **kwargs)

                return wrapper

            setattr(cls, name, mk_wrapper(name, meth))


_FakeRandomGen._delegate()


###################
# Random k-tuples #
###################


def ith_k_tuple(
    indices: NDArray[np.integer], /, *, n: int, k: int
) -> NDArray[np.int64]:
    """Calculate the k-tuple corresponding to the given :func:`itertools.product` index.

    Given the `n**k` possible k-tuples, this function returns the k-tuples that
    `[list(product(range(n), repeat=k))[i] for i in indices]` would evaluate to.

    Parameters
    ----------
    indices
        The tuple indices (must all be in `range(n**k)`).
    n
        The number of possible choices.
    k
        The length of each tuple.

    Returns
    -------
    A 2D array where each row is the k-tuple corresponding to the index.

    Raises
    ------
    ValueError
        If any of the indices are out of range.
    """
    if np.any((indices < 0) | (indices >= n**k)):  # pragma: no cover
        msg = f"Indices are out of range({n**k})."
        raise ValueError(msg)

    power_of_n = n ** np.arange(k - 1, -1, -1)  # [n^(k-1), n^(k-2), ..., n^0]
    return (indices[:, None] // power_of_n) % n


def random_k_tuples(
    size: int, *, n: int, k: int, rng: SeedLike | RNGLike | None = None
) -> NDArray[np.int64]:
    """Draw `size` distinct `k`-tuples of values in `range(n)` from the `n**k` possible tuples.

    Parameters
    ----------
    size
        The number of distinct k-tuples to draw.
    n
        The number of possible choices.
    k
        The length of the tuple.
    rng
        The random number generator to use.

    Returns
    -------
    A 2D array where each row is a distinct k-tuple.

    Raises
    ------
    ValueError
        If `size` is greater than the total number of possible k-tuples.
    """
    rng = np.random.default_rng(rng)

    total_possible = n**k
    if size > total_possible:  # pragma: no cover
        msg = f"Error: Cannot draw {size} unique k-tuples. Total possible is {total_possible}."
        raise ValueError(msg)

    indices = rng.choice(total_possible, size=size, replace=False)
    return ith_k_tuple(indices, n=n, k=k)


def random_str(
    size: int = 1, *, length: int, alphabet: str, rng: SeedLike | RNGLike | None = None
) -> NDArray[np.str_]:
    """Draw `size` distinct strings of length `k` from the given alphabet.

    Parameters
    ----------
    size
        The number of distinct strings to draw.
    length
        The length of each string.
    alphabet
        The alphabet to draw from.
    rng
        The random number generator to use.

    Returns
    -------
    A 0-1D array where each element is a distinct string of length `length`.
    """
    letters = np.array(list(alphabet), dtype="U1")
    indices = random_k_tuples(size, n=len(letters), k=length, rng=rng)
    return letters[indices].view(f"U{length}").squeeze()



"""Plotting functions and classes."""

from __future__ import annotations

from .._compat import deprecated
from . import palettes
from ._anndata import (
    clustermap,
    correlation_matrix,
    dendrogram,
    heatmap,
    ranking,
    scatter,
    tracksplot,
    violin,
)
from ._dotplot import DotPlot, dotplot
from ._matrixplot import MatrixPlot, matrixplot
from ._preprocessing import filter_genes_dispersion, highly_variable_genes
from ._qc import highest_expr_genes
from ._rcmod import set_rcParams_defaults, set_rcParams_scanpy
from ._scrublet import scrublet_score_distribution
from ._stacked_violin import StackedViolin, stacked_violin
from ._tools import (
    dpt_groups_pseudotime,
    dpt_timeseries,
    embedding_density,
    pca_loadings,
    pca_overview,
    pca_scatter,
    pca_variance_ratio,
    rank_genes_groups,
    rank_genes_groups_dotplot,
    rank_genes_groups_heatmap,
    rank_genes_groups_matrixplot,
    rank_genes_groups_stacked_violin,
    rank_genes_groups_tracksplot,
    rank_genes_groups_violin,
    sim,
)
from ._tools.paga import (
    paga,
    paga_adjacency,  # noqa: F401
    paga_compare,
    paga_path,
)
from ._tools.scatterplots import (
    diffmap,
    draw_graph,
    embedding,
    pca,
    spatial,
    tsne,
    umap,
)
from ._utils import matrix, timeseries, timeseries_as_heatmap, timeseries_subplot

__all__ = [
    "DotPlot",
    "MatrixPlot",
    "StackedViolin",
    "clustermap",
    "correlation_matrix",
    "dendrogram",
    "diffmap",
    "dotplot",
    "dpt_groups_pseudotime",
    "dpt_timeseries",
    "draw_graph",
    "embedding",
    "embedding_density",
    "filter_genes_dispersion",
    "heatmap",
    "highest_expr_genes",
    "highly_variable_genes",
    "matrix",
    "matrixplot",
    "paga",
    "paga_compare",
    "paga_path",
    "palettes",
    "pca",
    "pca_loadings",
    "pca_overview",
    "pca_scatter",
    "pca_variance_ratio",
    "rank_genes_groups",
    "rank_genes_groups_dotplot",
    "rank_genes_groups_heatmap",
    "rank_genes_groups_matrixplot",
    "rank_genes_groups_stacked_violin",
    "rank_genes_groups_tracksplot",
    "rank_genes_groups_violin",
    "ranking",
    "scatter",
    "scrublet_score_distribution",
    "set_rcParams_defaults",
    "set_rcParams_scanpy",
    "sim",
    "spatial",
    "stacked_violin",
    "tracksplot",
    "tsne",
    "umap",
    "violin",
]

timeseries = deprecated("Use `dpt_timeseries`.")(timeseries)
timeseries_as_heatmap = deprecated("Use `dpt_timeseries`.")(timeseries_as_heatmap)
timeseries_subplot = deprecated("Use `dpt_timeseries`.")(timeseries_subplot)



from __future__ import annotations

import scanpy as sc
import scanpy.external as sce
from testing.scanpy._helpers.data import pbmc3k
from testing.scanpy._pytest.marks import needs

pytestmark = [needs.wishbone]


def test_run_wishbone():
    adata = pbmc3k()
    sc.pp.normalize_per_cell(adata)
    sc.pp.neighbors(adata, n_pcs=15, n_neighbors=10)
    sc.pp.pca(adata)
    sc.tl.tsne(adata=adata, n_pcs=5, perplexity=30)
    sc.tl.diffmap(adata, n_comps=10)

    sce.tl.wishbone(
        adata=adata,
        start_cell="ACAAGAGACTTATC-1",
        components=[2, 3],
        num_waypoints=150,
    )
    assert all(k in adata.obs for k in ["trajectory_wishbone", "branch_wishbone"]), (
        "Run Wishbone Error!"
    )



"""Metrics which don't quite deserve their own file."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from natsort import natsorted
from pandas.api.types import CategoricalDtype

if TYPE_CHECKING:
    from collections.abc import Sequence


def confusion_matrix(
    orig: pd.Series | np.ndarray | Sequence,
    new: pd.Series | np.ndarray | Sequence,
    data: pd.DataFrame | None = None,
    *,
    normalize: bool = True,
) -> pd.DataFrame:
    """Given an original and new set of labels, create a labelled confusion matrix.

    Parameters `orig` and `new` can either be entries in data or categorical arrays
    of the same size.

    Params
    ------
    orig
        Original labels.
    new
        New labels.
    data
        Optional dataframe to fill entries from.
    normalize
        Should the confusion matrix be normalized?


    Examples
    --------

    .. plot::

        import scanpy as sc; import seaborn as sns
        pbmc = sc.datasets.pbmc68k_reduced()
        cmtx = sc.metrics.confusion_matrix("bulk_labels", "louvain", pbmc.obs)
        sns.heatmap(cmtx)

    """
    from sklearn.metrics import confusion_matrix as _confusion_matrix

    if data is not None:
        if isinstance(orig, str):
            orig = data[orig]
        if isinstance(new, str):
            new = data[new]

    # Coercing so I don't have to deal with it later
    orig, new = pd.Series(orig), pd.Series(new)
    assert len(orig) == len(new)

    unique_labels = pd.unique(np.concatenate((orig.values, new.values)))

    # Compute
    mtx = _confusion_matrix(orig, new, labels=unique_labels)
    if normalize:
        sums = mtx.sum(axis=1)[:, np.newaxis]
        mtx = np.divide(mtx, sums, where=sums != 0)

    # Label
    orig_name = "Original labels" if orig.name is None else orig.name
    new_name = "New Labels" if new.name is None else new.name
    df = pd.DataFrame(
        mtx,
        index=pd.Index(unique_labels, name=orig_name),
        columns=pd.Index(unique_labels, name=new_name),
    )

    # Filter
    if isinstance(orig.dtype, CategoricalDtype):
        orig_idx = pd.Series(orig).cat.categories
    else:
        orig_idx = natsorted(pd.unique(orig))
    if isinstance(new.dtype, CategoricalDtype):
        new_idx = pd.Series(new).cat.categories
    else:
        new_idx = natsorted(pd.unique(new))
    df = df.loc[np.array(orig_idx), np.array(new_idx)]

    return df
